import enum


class DispatchType(enum.Enum):
    GENERATOR = "GENERATOR"
    LOAD = "LOAD"


def dispatch_type_string(dispatch_type: DispatchType) -> str:
    if dispatch_type == DispatchType.GENERATOR:
        return "GENERATOR"

    return "LOAD"


def parse_dispatch_type(dispatch_string: str):
    """
        Converts dispatch type string into an enum constant

    """
    if not dispatch_string:
        return None

    dispatch_string = dispatch_string.lower().strip()

    if dispatch_string == "load":
        return DispatchType.LOAD

    if dispatch_string == "generating":
        return DispatchType.GENERATOR

    if dispatch_string == "generator":
        return DispatchType.GENERATOR

    raise Exception("Unknown dispatch type: {}".format(dispatch_string))



from datetime import datetime
from enum import Enum
from typing import List, Optional

from pydantic import Field

from opennem.api.schema import ApiBase


class StationIDListLocation(ApiBase):
    state: str
    country: str = "au"


class StationIDList(ApiBase):
    id: int
    name: str
    location: StationIDListLocation


class StationRecord(ApiBase):
    id: int

    code: str

    name: Optional[str]

    # Original network fields
    network_name: Optional[str]

    location_id: Optional[int]

    approved: bool = False

    # network: Optional[NetworkSchema] = None

    description: Optional[str]
    wikipedia_link: Optional[str]
    wikidata_id: Optional[str]

    created_by: Optional[str]
    created_at: Optional[datetime]


class StationUpdateResponse(ApiBase):
    success: bool = False
    record: StationRecord


class StationResponse(ApiBase):
    success: bool = False
    record_num: int
    records: Optional[List[StationRecord]]


class StationModificationTypes(str, Enum):
    approve = "approve"
    reject = "reject"


class StationModification(ApiBase):
    comment: Optional[str] = Field(None)
    modification: StationModificationTypes



"""
    Map network regions to BOM stations
"""

from typing import Optional

from opennem.core.loader import load_data

NETWORK_REGION_BOM_STATION = load_data("network_region_bom_station_map.json")


def get_network_region_weather_station(region_code: str) -> Optional[str]:
    """
    Maps a network region to a weather station id

    """

    region_code = region_code.upper()

    if not type(NETWORK_REGION_BOM_STATION) is dict:
        raise Exception("NETWORK_REGION_BOM_STATION invalid data type")

    if region_code in NETWORK_REGION_BOM_STATION:
        return NETWORK_REGION_BOM_STATION[region_code]

    return None



import logging

import click
from scrapy.utils.python import garbage_collect

from opennem.api.export.map import PriorityType
from opennem.api.export.tasks import export_all_monthly, export_energy, export_power
from opennem.db.load_fixtures import load_bom_stations_json, load_fixtures
from opennem.db.tasks import refresh_views
from opennem.db.views import init_aggregation_policies
from opennem.db.views.init import init_views_cli
from opennem.importer.all import run_all
from opennem.importer.db import import_facilities
from opennem.importer.db import init as db_init
from opennem.importer.emissions import import_emissions_map
from opennem.importer.mms import mms_export
from opennem.importer.opennem import opennem_export, opennem_import
from opennem.settings import settings
from opennem.workers.energy import run_energy_update_archive

logger = logging.getLogger("opennem.cli")


@click.group()
def main() -> None:
    pass


@click.command()
def crawl() -> None:
    logger.info("crawl @TODO")


@click.group()
def cmd_db() -> None:
    pass


@click.command()
def cmd_db_init() -> None:
    db_init()


@click.command()
def cmd_db_fixturer() -> None:
    load_fixtures()


@click.command()
def cmd_db_refresh() -> None:
    refresh_views()


@click.command()
@click.option("--purge", is_flag=True, help="Purge unmapped views")
def cmd_db_views(purge: bool) -> None:
    init_views_cli(purge=purge)


@click.command()
def cmd_db_aggregations() -> None:
    init_aggregation_policies()


@click.group()
def cmd_import() -> None:
    pass


@click.group()
def cmd_export() -> None:
    pass


@click.command()
def cmd_import_opennem() -> None:
    opennem_import()


@click.command()
def cmd_import_facilities() -> None:
    import_facilities()


@click.command()
def cmd_import_mms() -> None:
    mms_export()


@click.command()
def cmd_import_all() -> None:
    run_all()


@click.command()
def cmd_import_emissions() -> None:
    import_emissions_map("emission_factors.csv")


@click.command()
def cmd_export_opennem() -> None:
    opennem_export()


@click.command()
def cmd_export_power() -> None:
    export_power(priority=PriorityType.live)


@click.command()
def cmd_export_energy() -> None:
    export_energy()


@click.command()
def cmd_export_energy_monthly() -> None:
    export_all_monthly()


@click.command()
def cmd_export_all() -> None:
    run_all()


@click.group()
def cmd_weather() -> None:
    pass


@click.command()
def cmd_weather_init() -> None:
    load_bom_stations_json()


@click.group()
def cmd_task() -> None:
    pass


@click.command()
@click.option("--year", required=True, type=int)
def cmd_task_energy(year: int) -> None:
    run_energy_update_archive(year)


main.add_command(crawl)
main.add_command(cmd_db, name="db")
main.add_command(cmd_import, name="import")
main.add_command(cmd_export, name="export")
main.add_command(cmd_weather, name="weather")
main.add_command(cmd_task, name="task")

cmd_import.add_command(cmd_import_opennem, name="opennem")
cmd_import.add_command(cmd_import_mms, name="mms")
cmd_import.add_command(cmd_import_all, name="all")
cmd_import.add_command(cmd_import_emissions, name="emissions")
cmd_import.add_command(cmd_import_facilities, name="facilities")

cmd_export.add_command(cmd_export_opennem, name="opennem")
cmd_export.add_command(cmd_export_all, name="all")
cmd_export.add_command(cmd_export_power, name="power")
cmd_export.add_command(cmd_export_energy, name="energy")
cmd_export.add_command(cmd_export_energy_monthly, name="energy_monthly")

cmd_db.add_command(cmd_db_init, name="init")
cmd_db.add_command(cmd_db_fixturer, name="load_fixtures")
cmd_db.add_command(cmd_db_refresh, name="refresh")
cmd_db.add_command(cmd_db_views, name="views")
cmd_db.add_command(cmd_db_aggregations, name="aggregations")

cmd_weather.add_command(cmd_weather_init, name="init")

cmd_task.add_command(cmd_task_energy, name="energy")

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        logger.error("User interrupted")
    except Exception as e:
        logger.error(e)

        if settings.debug:
            import traceback

            traceback.print_exc()

    finally:
        garbage_collect()



# pylint: disable=no-member
"""
Fix network ids

Revision ID: d886acb494ca
Revises: eeb81f56767e
Create Date: 2020-11-24 11:48:22.720354

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "d886acb494ca"
down_revision = "eeb81f56767e"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        "update facility set network_id='NEM' where network_region in ('NSW1', 'QLD1', 'SA1', 'TAS1', 'VIC1')"
    )
    op.execute(
        "update facility set network_id='WEM' where network_region in ('WEM')"
    )


def downgrade():
    pass



import logging
from datetime import datetime
from io import BytesIO
from typing import Dict, Generator
from zipfile import ZipFile

import scrapy

from opennem.pipelines.bulk_insert import BulkInsertPipeline
from opennem.pipelines.csv import RecordsToCSVPipeline
from opennem.pipelines.nem import ExtractCSV
from opennem.pipelines.nem.opennem import NemwebUnitScadaOpenNEMStorePipeline
from opennem.utils.dates import get_date_component, month_series
from opennem.utils.handlers import _handle_zip, chain_streams
from opennem.utils.mime import decode_bytes, mime_from_content, mime_from_url

logger = logging.getLogger(__name__)

MMS_URL = "http://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM/{year}/MMSDM_{year}_{month}/MMSDM_Historical_Data_SQLLoader/DATA/PUBLIC_DVD_{table}_{year}{month}010000.zip"


class MMSArchiveBulkSpider(scrapy.Spider):
    name = "au.mms.archive.dispatch_scada"

    tables = ["DISPATCH_UNIT_SCADA"]

    pipelines = set(
        [
            ExtractCSV,
            NemwebUnitScadaOpenNEMStorePipeline,
            BulkInsertPipeline,
            RecordsToCSVPipeline,
        ]
    )

    def start_requests(self) -> Generator[scrapy.Request, None, None]:
        start_month = datetime(2020, 1, 1)
        end_month = datetime(2009, 7, 1)

        for date in month_series(start_month, end_month):
            for table in self.tables:
                url_params = {
                    "month": get_date_component("%m", dt=date),
                    "year": get_date_component("%Y", dt=date),
                    "table": table.upper(),
                }

                req_url = MMS_URL.format(**url_params)

                yield scrapy.Request(req_url)

    def parse(self, response) -> Generator[Dict, None, None]:
        content = None

        file_mime = mime_from_content(response.body)

        if not file_mime:
            file_mime = mime_from_url(response.url)

        if file_mime == "application/zip":
            with ZipFile(BytesIO(response.body)) as zf:
                if len(zf.namelist()) == 1:
                    content = zf.open(zf.namelist()[0]).read()

                c = []
                stream_count = 0

                for filename in zf.namelist():
                    if filename.endswith(".zip"):
                        c.append(_handle_zip(zf.open(filename), "r"))
                        stream_count += 1
                    else:
                        c.append(zf.open(filename))

                content = chain_streams(c).read()
        else:
            content = response.body.getvalue()

        if not content:
            logger.info("No content from scrapy request")
            return None

        content = decode_bytes(content)

        item = {}
        item["content"] = content
        item["extension"] = ".csv"
        item["mime_type"] = file_mime

        yield item


class MMSArchivePriceSpider(MMSArchiveBulkSpider):
    name = "au.mms.archive.dispatch_price"

    pipelines = set(
        [
            ExtractCSV,
            NemwebUnitScadaOpenNEMStorePipeline,
        ]
    )

    tables = ["DISPATCHPRICE"]


class MMSArchiveDispatchISSpider(MMSArchiveBulkSpider):
    name = "au.mms.archive.dispatch_is"

    pipelines = set(
        [
            ExtractCSV,
            NemwebUnitScadaOpenNEMStorePipeline,
        ]
    )

    tables = ["DISPATCHREGIONSUM"]


class MMSArchiveRooftopActualSpider(MMSArchiveBulkSpider):
    name = "au.mms.archive.rooftop_actual"

    tables = ["ROOFTOP_PV_ACTUAL"]


class MMSArchiveTradingPriceSpider(MMSArchiveBulkSpider):
    name = "au.mms.archive.trading_price"

    pipelines = set(
        [
            ExtractCSV,
            NemwebUnitScadaOpenNEMStorePipeline,
        ]
    )

    tables = ["TRADINGPRICE"]



import json
from functools import wraps
from typing import Any, Callable, Optional, Type

from fastapi_cache import FastAPICache
from fastapi_cache.coder import Coder

from opennem.exporter.encoders import OpenNEMJSONEncoder


class PydanticCoder(Coder):
    @classmethod
    def encode(cls, value: Any) -> str:
        return json.dumps(value, cls=OpenNEMJSONEncoder)

    @classmethod
    def decode(cls, value: Any) -> Any:
        return json.loads(value, cls=OpenNEMJSONEncoder)


def cache(
    expire: int = None,
    coder: Type[Coder] = None,
    key_builder: Callable = None,
    namespace: Optional[str] = "",
):
    """
    cache all function
    :param namespace:
    :param expire:
    :param coder:
    :param key_builder:
    :return:
    """

    def wrapper(func):
        @wraps(func)
        async def inner(*args, **kwargs):
            nonlocal coder
            nonlocal expire
            nonlocal key_builder
            copy_kwargs = kwargs.copy()
            request = copy_kwargs.pop("request", None)
            response = copy_kwargs.pop("response", None)
            if request and request.headers.get("Cache-Control") == "no-store":
                return func(*args, **kwargs)

            coder = coder or FastAPICache.get_coder()
            expire = expire or FastAPICache.get_expire()
            key_builder = key_builder or FastAPICache.get_key_builder()
            backend = FastAPICache.get_backend()

            cache_key = key_builder(
                func, namespace, request=request, response=response, args=args, kwargs=copy_kwargs
            )
            ttl, ret = await backend.get_with_ttl(cache_key)
            if not request:
                if ret is not None:
                    return coder.decode(ret)
                ret = func(*args, **kwargs)
                await backend.set(
                    cache_key, coder.encode(ret), expire or FastAPICache.get_expire()
                )
                return ret

            if request.method != "GET":
                return func(request, *args, **kwargs)
            if_none_match = request.headers.get("if-none-match")
            if ret is not None:
                if response:
                    response.headers["Cache-Control"] = f"max-age={ttl}"
                    etag = f"W/{hash(ret)}"
                    if if_none_match == etag:
                        response.status_code = 304
                        return response
                    response.headers["ETag"] = etag
                return coder.decode(ret)

            ret = await func(*args, **kwargs)
            await backend.set(cache_key, coder.encode(ret), expire or FastAPICache.get_expire())
            return ret

        return inner

    return wrapper



"""
    HTTP module with custom timeout and retry adaptors

    usage:

    from opennem.utils.http import http
    http.get(`url`) etc.

"""
import logging

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from opennem.settings import settings

logger = logging.getLogger("opennem.utils.http")

DEFAULT_TIMEOUT = settings.http_timeout
DEFAULT_RETRIES = settings.http_retries


def setup_http_cache() -> bool:
    """Sets up requests session local cachine using
    requests-cache if enabled in settings"""
    if not settings.http_cache_local:
        return False

    try:
        import requests_cache
    except ImportError:
        logger.error("Request caching requires requests-cache library")
        return False

    requests_cache.install_cache(".opennem_requests_cache", expire_after=300)
    return True


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.timeout = DEFAULT_TIMEOUT
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


retry_strategy = Retry(
    total=DEFAULT_RETRIES,
    backoff_factor=1,
    status_forcelist=[403, 429, 500, 502, 503, 504],
    method_whitelist=["HEAD", "GET", "OPTIONS"],
)


http = requests.Session()

adapter_timeout = TimeoutHTTPAdapter()
http.mount("https://", adapter_timeout)
http.mount("http://", adapter_timeout)


adapter_retry = HTTPAdapter(max_retries=retry_strategy)
http.mount("https://", adapter_retry)
http.mount("http://", adapter_retry)

setup_http_cache()



import logging
from datetime import datetime, timedelta
from typing import Dict, List, Optional

from pytz import FixedOffset

from opennem.api.stats.controllers import get_scada_range
from opennem.api.time import human_to_interval, human_to_period
from opennem.core.energy import energy_sum
from opennem.db import get_database_engine
from opennem.db.models.opennem import FacilityScada
from opennem.diff.versions import CUR_YEAR, get_network_regions
from opennem.notifications.slack import slack_message
from opennem.pipelines.bulk_insert import build_insert_query
from opennem.pipelines.csv import generate_csv_from_records
from opennem.schema.dates import DatetimeRange, TimeSeries
from opennem.schema.network import NetworkNEM, NetworkSchema

logger = logging.getLogger("opennem.workers.energy")

DRY_RUN = False


def get_generated(
    network_region: str,
    date_min: datetime,
    date_max: datetime,
    network: NetworkSchema,
    fueltech_id: Optional[str] = None,
) -> List[Dict]:
    """Gets generated values for a date range for a network and network region
    and optionally for a single fueltech"""

    # @TODO support refresh energies for a single duid or station

    __sql = """
    select
        fs.trading_interval at time zone 'AEST' as trading_interval,
        fs.facility_code,
        fs.network_id,
        generated
    from
        facility_scada fs
        left join facility f on fs.facility_code = f.code
    where
        fs.network_id='{network_id}'
        and f.network_region='{network_region}'
        and fs.trading_interval >= '{date_min}'
        and fs.trading_interval <= '{date_max}'
        and fs.is_forecast is False
        and f.fueltech_id not in ('solar_rooftop', 'imports', 'exports')
        and f.interconnector is False
        {fueltech_match}
    order by fs.trading_interval asc, 2;
    """

    fueltech_match = ""

    if fueltech_id:
        fueltech_match = f"and f.fueltech_id = '{fueltech_id}'"

    query = __sql.format(
        network_id=network.code,
        network_region=network_region,
        date_min=date_min,
        date_max=date_max,
        fueltech_match=fueltech_match,
    )

    engine = get_database_engine()

    results = []

    with engine.connect() as c:
        logger.debug(query)

        if not DRY_RUN:
            try:
                results = list(c.execute(query))
            except Exception as e:
                logger.error(e)

    logger.debug("Got back {} rows".format(len(results)))

    return results


def insert_energies(results: List[Dict]) -> int:
    """Takes a list of generation values and calculates energies and bulk-inserts
    into the database"""

    # Get the energy sums as a dataframe
    esdf = energy_sum(results)

    # Add metadata
    esdf["created_by"] = "opennem.worker.energy"
    esdf["created_at"] = ""
    esdf["updated_at"] = datetime.now()
    esdf["generated"] = None
    esdf["is_forecast"] = False

    # reorder columns
    columns = [
        "created_by",
        "created_at",
        "updated_at",
        "network_id",
        "trading_interval",
        "facility_code",
        "generated",
        "eoi_quantity",
        "is_forecast",
    ]
    esdf = esdf[columns]

    records_to_store: List[Dict] = esdf.to_dict("records")

    # Build SQL + CSV and bulk-insert
    sql_query = build_insert_query(FacilityScada, ["updated_at", "eoi_quantity"])
    conn = get_database_engine().raw_connection()
    cursor = conn.cursor()

    csv_content = generate_csv_from_records(
        FacilityScada,
        records_to_store,
        column_names=records_to_store[0].keys(),
    )

    cursor.copy_expert(sql_query, csv_content)
    conn.commit()

    logger.info("Inserted {} records".format(len(records_to_store)))

    return len(records_to_store)


def get_date_range(network: NetworkSchema) -> DatetimeRange:
    date_range = get_scada_range(network=NetworkNEM)

    time_series = TimeSeries(
        start=date_range.start,
        end=date_range.end,
        interval=human_to_interval("1d"),
        period=human_to_period("all"),
        network=network,
    )

    return time_series.get_range()


def run_energy_calc(
    region: str,
    date_min: datetime,
    date_max: datetime,
    network: NetworkSchema,
    fueltech_id: Optional[str] = None,
) -> int:
    results = get_generated(region, date_min, date_max, network=network, fueltech_id=fueltech_id)
    num_records = 0

    try:
        if len(results) < 1:
            raise Exception("No results from get_generated query")

        num_records = insert_energies(results)
        logger.info("Done {} for {} => {}".format(region, date_min, date_max))
    except Exception as e:
        logger.error(e)
        slack_message("Energy archive error: {}".format(e))

    return num_records


def run_energy_update_archive(
    year: int = 2021,
    months: Optional[List[int]] = None,
    days: Optional[int] = None,
    regions: Optional[List[str]] = None,
    fueltech_id: Optional[str] = None,
    network: NetworkSchema = NetworkNEM,
) -> None:

    date_range = get_date_range(network=network)

    if not months:
        months = list(range(1, 12))

    if not regions:
        regions = [i.code for i in get_network_regions(network)]

    for month in months:
        date_min = datetime(
            year=year, month=month, day=1, hour=0, minute=0, second=0, tzinfo=FixedOffset(600)
        )

        date_max = datetime(
            year=year,
            month=month + 1,
            day=1,
            hour=0,
            minute=0,
            second=0,
            tzinfo=FixedOffset(600),
        )

        if days:
            date_max = datetime(
                year=year,
                month=month,
                day=1 + days,
                hour=0,
                minute=0,
                second=0,
                tzinfo=FixedOffset(600),
            )

        date_max = date_max + timedelta(minutes=5)

        if date_max > date_range.end:
            date_max = date_range.end

        if date_min > date_max:
            slack_message("Reached end of energy archive")
            logger.debug("reached end of archive")
            break

        for region in regions:
            run_energy_calc(region, date_min, date_max, fueltech_id=fueltech_id, network=network)


def run_energy_update_yesterday(
    network: NetworkSchema = NetworkNEM,
) -> None:
    """Run energy sum update for yesterday. This task is scheduled
    in scheduler/db

    This is NEM only atm"""

    # today_midnight in NEM time
    today_midnight = datetime.now().replace(
        tzinfo=network.get_fixed_offset(), microsecond=0, hour=0, minute=0, second=0
    )

    date_min = today_midnight - timedelta(days=1)
    date_max = date_min + timedelta(days=1, minutes=5)

    regions = [i.code for i in get_network_regions(network)]

    for region in regions:
        run_energy_calc(region, date_min, date_max, network=network)

    slack_message("Ran energy dailies for regions: {}".format(",".join(regions)))


def run_energy_update_all() -> None:
    """Runs energy update for all regions and all years for one-off
    inserts"""
    for year in range(2011, CUR_YEAR + 1):
        run_energy_update_archive(year=year)


if __name__ == "__main__":
    run_energy_update_yesterday()



from opennem.spiders.wem import WemHistoricSpider


class WemHistoricLoadSummary(WemHistoricSpider):
    name = "au.wem.archive.load_summary"
    start_urls = (
        "http://data.wa.aemo.com.au/public/public-data/datafiles/load-summary/"
    )
    limit = 0



"""
    Take records generated from previous pipeline steps (usually a list of
    dicts) and prep them into CSV in the shape of the database table
    schema so they can be used with bulk_insert

"""
import csv
import logging
from io import StringIO
from typing import Dict, List, Optional, Union

from scrapy import Spider
from sqlalchemy.sql.schema import Table

from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


def generate_csv_from_records(
    table: Table, records: List[Dict], column_names: Optional[List[str]] = None
) -> StringIO:
    """
    Take a list of dict records and a table schema and generate a csv
    buffer to be used in bulk_insert

    """
    if len(records) < 1:
        raise Exception("No records")

    csv_buffer = StringIO()

    table_column_names = [c.name for c in table.__table__.columns.values()]  # type: ignore

    # sanity check the records we received to make sure
    # they match the table schema
    if isinstance(records, list):
        first_record = records[0]

        record_field_names = list(first_record.keys())

        for field_name in record_field_names:
            if field_name not in table_column_names:
                raise Exception("Column name not found: {}".format(field_name))

        for column_name in table_column_names:
            if column_name not in record_field_names:
                raise Exception(
                    "Missing value for column {}".format(column_name)
                )

        column_names = record_field_names

    if not column_names:
        column_names = table_column_names

    csvwriter = csv.DictWriter(csv_buffer, fieldnames=column_names)
    csvwriter.writeheader()

    # @TODO put the columns in the correct order ..
    # @NOTE do we need to ?!
    for record in records:
        csvwriter.writerow(record)

    # rewind it back to the start
    csv_buffer.seek(0)

    return csv_buffer


class RecordsToCSVPipeline(object):
    """
    Pipeline that adapts the CSV generator and bulk inserter

    """

    @check_spider_pipeline
    def process_item(
        self, item: Union[List[Dict], Dict], spider: Optional[Spider] = None
    ) -> List[Dict]:

        # if it's a single item put it in a list
        if not isinstance(item, list):
            item = [item]

        for record_set in item:
            if not isinstance(record_set, dict):
                logger.error(
                    "Invalid record_set passed to CSV pipeline: %s", record_set
                )
                return record_set

            if "table_schema" not in record_set:
                logger.error("No table model passed to csv generator")
                logger.error(record_set)
                return item

            table = record_set["table_schema"]

            if "records" not in record_set:
                logger.error("No records to generate csv from")
                return item

            records = record_set["records"]

            csv_content = generate_csv_from_records(table, records)

            record_set["csv"] = csv_content

        return item



"""
OpenNEM v2 -> v3 compatability methods

"""

from opennem.core.fueltechs import map_v2_fueltech, map_v3_fueltech
from opennem.schema.network import NetworkNEM


def translate_id_v3_to_v2(v3_id: str) -> str:
    """Translates v3 ids to v2 ids"""
    idc = id2 = v3_id.split(".")

    if idc[0].lower() == "nem":
        id2 = id2[1:]

    stat_type = idc[-1]

    if stat_type in ["energy", "power", "market_value", "emissions"]:
        fueltech_old = idc[-2]
        fueltech = map_v3_fueltech(fueltech_old)
        id2 = [idc[1], "fuel_tech", fueltech, stat_type]

    if stat_type in ["temperature", "temperature_mean", "temperature_max"]:
        id2 = [idc[1], idc[-1]]

    return ".".join([i for i in id2 if i])


def translate_id_v2_to_v3(v2_id: str) -> str:
    network = NetworkNEM

    idc = v2_id.split(".")

    if idc[0] != network.code:
        idc.insert(0, network.code.lower())

    stat_type = idc[-1]

    if stat_type in ["energy", "power", "market_value", "emissions"]:
        fueltech_old = idc[-2]
        fueltech = map_v2_fueltech(fueltech_old)
        # id2 = [idc[1], "fuel_tech", fueltech, stat_type]
        idc[-2] = fueltech

    return ".".join([i for i in idc if i])


def fix_date_v2(date_str: str) -> str:
    """Fix v2 dates so they're ISO"""
    if date_str.endswith("+1000"):
        date_str = date_str[:-2]

    return date_str



import logging

from opennem.db.views import init_aggregation_policies, init_database_views, purge_views

logger = logging.getLogger("opennem.db.views")


def init_views_cli(purge: bool = False) -> None:
    logger.info("init views")

    if purge:
        purge_views()

    init_database_views()
    init_aggregation_policies()


if __name__ == "__main__":
    purge_views()
    init_database_views()
    init_aggregation_policies()



from __future__ import annotations

from datetime import date, datetime, timedelta
from decimal import Decimal
from math import isnan
from typing import Any, List, Optional, Tuple, Union

from pydantic import validator

from opennem.api.time import human_to_interval
from opennem.core.compat.utils import translate_id_v3_to_v2
from opennem.core.fueltechs import map_v3_fueltech
from opennem.schema.core import BaseConfig
from opennem.schema.network import NetworkSchema
from opennem.schema.time import TimeIntervalAPI, TimePeriodAPI
from opennem.settings import settings
from opennem.utils.interval import get_human_interval
from opennem.utils.numbers import sigfig_compact
from opennem.utils.timezone import get_current_timezone


def chop_microseconds(dt: datetime) -> datetime:
    if not dt.microsecond:
        return dt

    return dt - timedelta(microseconds=dt.microsecond)


def optionaly_lowercase_string(value: str) -> str:
    """Read from settings if we want output schema string
    values to be lowercased or not and perform"""
    if settings.schema_output_lowercase_strings:
        value = value.lower()

    return value


def number_output(n: Union[float, int, None]) -> Optional[Union[float, int, None, Decimal]]:
    """Format numbers for data series outputs"""
    if n is None:
        return None

    if n == 0:
        return 0

    if isnan(n):
        return None

    return sigfig_compact(n)


def data_validate(values: List[Union[float, int, None, Decimal]]) -> List[Union[float, int, None]]:
    """Validate and format list of numeric data values"""
    return list(
        map(
            number_output,
            values,
        )
    )


def optionally_parse_string_datetime(
    value: Optional[Union[str, datetime, date]] = None
) -> Optional[Union[str, datetime, date]]:
    if not value:
        return value

    if isinstance(value, str):
        return datetime.fromisoformat(value)

    return value


def get_data_id(
    network: NetworkSchema,
) -> str:
    """Method to build the data id for a stat export"""
    id_components = []

    if settings.schema_output_id_country:
        id_components.append(network.country)

    id_str = ".".join([i.lower() for i in id_components if i])

    return id_str


class OpennemDataHistory(BaseConfig):
    start: datetime
    last: datetime
    interval: str
    data: List

    # validators
    _data_valid = validator("data", allow_reuse=True, pre=True)(data_validate)

    def values(self) -> List[Tuple[datetime, float]]:
        interval_obj = get_human_interval(self.interval)
        interval_def = human_to_interval(self.interval)
        inclusive = False
        dt = self.start

        if interval_def.interval < 1440:
            inclusive = True

        # return as list rather than generate
        timeseries_data = []

        # rewind back one interval
        if inclusive:
            # dt -= interval_obj
            pass

        for v in self.data:
            timeseries_data.append((dt, v))
            dt = dt + interval_obj

        # @TODO do some sanity checking here
        # if dt != self.last:
        #     raise Exception(
        #         "Mismatch between start, last and interval size. Got {} and {}".format(
        #             dt, self.last
        #         )
        #     )

        return timeseries_data


class OpennemData(BaseConfig):
    id: Optional[str]
    type: Optional[str]
    fuel_tech: Optional[str]

    network: Optional[str]
    region: Optional[str]
    data_type: str
    code: str = ""
    units: str

    interval: Optional[TimeIntervalAPI]
    period: Optional[TimePeriodAPI]

    history: OpennemDataHistory
    forecast: Optional[OpennemDataHistory]

    # validators

    # conveniance methods
    def id_v2(self) -> str:
        return translate_id_v3_to_v2(self.id)

    def fueltech_v2(self) -> Optional[str]:
        if self.fuel_tech:
            return map_v3_fueltech(self.fuel_tech)

        return None


class OpennemDataSet(BaseConfig):
    type: Optional[str]
    version: Optional[str]
    network: Optional[str]
    code: str
    region: Optional[str]
    created_at: Optional[datetime]

    data: List[OpennemData]

    def append_set(self, subject_set: Optional[OpennemDataSet] = None):
        if not subject_set:
            return None

        if not subject_set.data:
            return None

        if not isinstance(subject_set.data, list):
            return None

        if len(subject_set.data):
            self.data += subject_set.data

        return None

    def get_id(self, id: str) -> Optional[OpennemData]:
        _ds = list(filter(lambda x: x.id == id, self.data))

        if len(_ds) < 1:
            return None

        return _ds.pop()

    # validators
    _version_fromstr = validator("created_at", allow_reuse=True, pre=True)(
        optionally_parse_string_datetime
    )

    _created_at_trim = validator("created_at", allow_reuse=True, pre=True)(chop_microseconds)
    _network_lowercase = validator("network", allow_reuse=True, pre=True)(
        optionaly_lowercase_string
    )


class RegionFlowResult(BaseConfig):
    interval: datetime
    flow_from: str
    flow_to: str
    generated: Optional[float]
    flow_from_energy: Optional[float]
    flow_to_energy: Optional[float]


class RegionFlowEmissionsResult(BaseConfig):
    interval: datetime
    flow_from: str
    flow_to: str
    energy: float
    flow_from_emissions: Optional[float]
    flow_to_emissions: Optional[float]
    flow_from_intensity: Optional[float]
    flow_to_intensity: Optional[float]


class DataQueryResult(BaseConfig):
    interval: datetime
    result: Union[float, int, None, Decimal]
    group_by: Optional[str]


class ScadaDateRange(BaseConfig):
    start: datetime
    end: datetime
    network: Optional[NetworkSchema]

    def _get_value_localized(self, field_name: str = "start") -> Any:
        timezone = get_current_timezone()
        date_aware = getattr(self, field_name)

        if self.network:
            timezone = self.network.get_timezone()

        date_aware = date_aware.astimezone(timezone)

        return date_aware

    def get_start_year(self) -> int:
        return self.start.year

    def get_start(self) -> datetime:
        return self._get_value_localized("start")

    def get_end(self) -> datetime:
        return self._get_value_localized("end")

    def get_start_sql(self, as_date: bool = False) -> str:
        return "'{}'".format(self.get_start() if not as_date else self.get_start().date())

    def get_end_sql(self, as_date: bool = False) -> str:
        return "'{}'".format(self.get_end() if not as_date else self.get_end().date())



# pylint: disable=no-member
"""
Data fix price data

Revision ID: 156aec569e8d
Revises: a6261453b8e7
Create Date: 2021-01-31 00:11:27.874617

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "156aec569e8d"
down_revision = "a6261453b8e7"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute("update balancing_summary set price=NULL where network_id='NEM';")


def downgrade() -> None:
    pass



import json
import logging
import os

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker

from opennem.core import load_data_json
from opennem.db import db_connect
from opennem.db.models.opennem import Facility

logger = logging.getLogger(__name__)


def load_facility_fueltech_map():
    """
        Only do this for DUIDs that don't have a fueltech


    """

    engine = db_connect()
    session = sessionmaker(bind=engine)
    s = session()

    facility_fueltech_map = load_data_json("facility_fueltech_map.json")

    update_count = 0

    for facility_duid, facility_fueltech in facility_fueltech_map.items():

        facilities = (
            s.query(Facility)
            .filter(Facility.network_code == facility_duid)
            .all()
        )

        for f in facilities:
            if f.fueltech_id:
                logger.debug(
                    "Skipping {} because it already has a fueltech of {}".format(
                        f.code, f.fueltech_id
                    )
                )
                continue

            f.fueltech_id = facility_fueltech
            s.add(f)
            update_count += 1

    s.commit()
    logger.info("Updated {} fueltechs for facilities".format(update_count))


if __name__ == "__main__":
    load_facility_fueltech_map()



# pylint: disable=no-member
"""
Create mv_nem_facility_power_30min

Revision ID: ec9daa4f407c
Revises: 66e98f534a41
Create Date: 2021-01-26 01:05:05.043919

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "ec9daa4f407c"
down_revision = "66e98f534a41"
branch_labels = None
depends_on = None

stmt_drop = "drop materialized view if exists mv_nem_facility_power_30min cascade;"


stmt = """
CREATE MATERIALIZED VIEW mv_nem_facility_power_30min WITH (timescaledb.continuous) AS
select
    time_bucket('30 minutes', fs.trading_interval) as trading_interval,
    fs.facility_code,
    avg(fs.generated) as facility_power,
    fs.is_forecast
from facility_scada fs
where fs.network_id = 'NEM' and trading_interval > '2021-01-01 00:00:00+10'
group by
    1,
    fs.facility_code,
    fs.is_forecast;

"""


def upgrade() -> None:
    with op.get_context().autocommit_block():
        op.execute(stmt_drop)
        op.execute(stmt)


def downgrade() -> None:
    op.execute(stmt_drop)



"""


"""
from typing import List, Optional

from opennem.schema.core import BaseConfig

from .utils import format_offset


class ContinuousAggregationPolicy(BaseConfig):
    """Defines a continuous aggregation policy as used by
    timescale db. The map will destroy/create these"""

    interval: str
    start_interval: Optional[str]
    end_interval: Optional[str]

    @property
    def schedule_interval(self) -> str:
        return format_offset(self.interval)

    @property
    def start_offset(self) -> str:
        if self.start_interval:
            return format_offset(self.start_interval)

        return "NULL"

    @property
    def end_offset(self) -> str:
        if self.end_interval:
            return format_offset(self.end_interval)

        return "NULL"


class ViewDefinition(BaseConfig):
    """Defines a view (very useful docstrings!)"""

    # Priority determines order views
    # are processed in. probably better
    # done with a depends_on but eyyyy
    priority: int = 0

    name: str

    materialized: bool = False

    filepath: str

    view_content: Optional[str]

    # Creates a unique index
    primary_key: Optional[List[str]]

    aggregation_policy: Optional[ContinuousAggregationPolicy]



from datetime import datetime
from typing import List, Optional

from fastapi import APIRouter, Depends, HTTPException, Query
from sqlalchemy.orm import Session
from starlette import status

from opennem.core.dispatch_type import DispatchType
from opennem.db import get_database_engine, get_database_session
from opennem.db.models.opennem import Facility, FuelTech, Location, Revision, Station
from opennem.schema.opennem import StationSchema

from .schema import StationIDList, StationModification, StationRecord, StationUpdateResponse

router = APIRouter()


@router.get(
    "/",
    response_model=List[StationRecord],
    description="Get a list of all stations",
    response_model_exclude_none=True,
)
def stations(
    session: Session = Depends(get_database_session),
    facilities_include: Optional[bool] = Query(False, description="Include facilities in records"),
    revisions_include: Optional[bool] = Query(False, description="Include revisions in records"),
    history_include: Optional[bool] = Query(False, description="Include history in records"),
    only_approved: Optional[bool] = Query(
        False, description="Only show approved stations not those pending"
    ),
    name: Optional[str] = None,
    limit: Optional[int] = None,
    page: int = 1,
) -> List[StationSchema]:
    stations = (
        session.query(Station)
        .join(Location)
        .enable_eagerloads(True)
        #
        # .group_by(Station.code)
        # .filter(Facility.fueltech_id.isnot(None))
        # .filter(Facility.status_id.isnot(None))
    )

    if facilities_include:
        stations = stations.outerjoin(Facility, Facility.station_id == Station.id).outerjoin(
            FuelTech, Facility.fueltech_id == FuelTech.code
        )

    if revisions_include:
        stations = stations.outerjoin(Revision, Revision.station_id == Station.id)

    if only_approved:
        stations = stations.filter(Station.approved == True)

    if name:
        stations = stations.filter(Station.name.like("%{}%".format(name)))

    stations = stations.order_by(
        # Facility.network_region,
        Station.name,
        # Facility.network_code,
        # Facility.code,
    )

    stations = stations.all()

    return stations


@router.get(
    "/",
    name="station lookup",
    description="""Lookup station by code or network_code. \
        Require one of code or network_code""",
    response_model=List[StationRecord],
    response_model_exclude_none=True,
)
def station_lookup(
    session: Session = Depends(get_database_session),
    station_code: Optional[str] = Query(None, description="Code of the station to lookup"),
    network_code: Optional[str] = Query(None, description="The network code to lookup"),
) -> List[StationSchema]:
    if not station_code and not network_code:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    station = session.query(Station)

    if station_code:
        station = station.filter_by(code=station_code)

    if network_code:
        station = station.filter_by(network_name=network_code)

    stations = station.all()

    if not stations:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return stations


@router.get(
    "/ids",
    response_model=List[StationIDList],
    description="Get a list of station ids for dropdowns",
    response_model_exclude_none=True,
)
def station_ids(
    session: Session = Depends(get_database_session),
    only_approved: Optional[bool] = Query(
        True, description="Only show approved stations not those pending"
    ),
) -> List[StationIDList]:
    stations = session.query(Station).join(Station.location)

    if only_approved:
        stations = stations.filter(Station.approved == True)

    stations = stations.order_by(
        Station.id,
    )

    stations = stations.all()

    return stations


@router.get(
    "/{station_code:path}",
    response_model=StationSchema,
    description="Get a single station by code",
    response_model_exclude_none=True,
)
def station(
    session: Session = Depends(get_database_session),
    engine=Depends(get_database_engine),
    station_code: str = None,
    only_generators: bool = Query(True, description="Show only generators"),
    power_include: Optional[bool] = Query(False, description="Include last week of power output"),
    revisions_include: Optional[bool] = Query(False, description="Include revisions in records"),
    history_include: Optional[bool] = Query(False, description="Include history in records"),
) -> StationSchema:

    station = (
        session.query(Station)
        .filter(Station.code == station_code)
        .filter(Facility.station_id == Station.id)
        .filter(~Facility.code.endswith("NL1"))
    )

    if only_generators:
        station = station.filter(Facility.dispatch_type == DispatchType.GENERATOR)

    station = station.one_or_none()

    if not station:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Station not found")

    if not station.facilities:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Station has no facilities",
        )

    station.network = station.facilities[0].network_id

    if revisions_include:
        revisions = session.query(Revision).all()

        station.revisions = list(
            filter(
                lambda rev: rev.schema == "station" and rev.code == station.code,
                revisions,
            )
        )

        for facility in station.facilities:
            facility.revisions = list(
                filter(
                    lambda rev: rev.schema == "facility" and rev.code == facility.code,
                    revisions,
                )
            )

    if power_include:
        pass

    if station.location and station.location.geom:
        __query = """
            select
                code,
                ST_Distance(l.geom, bs.geom, false) / 1000.0 as dist
            from bom_station bs, location l
            where
                l.id = {id}
                and bs.priority < 2
            order by dist
            limit 1
        """.format(
            id=station.location.id
        )

        result = []

        with engine.connect() as c:
            result = list(c.execute(__query))

            if len(result):
                station.location.weather_nearest = {
                    "code": result[0][0],
                    "distance": round(result[0][1], 2),
                }

    return station


@router.get(
    "/history",
    name="station all history",
    description="""Get history for all stations""",
    response_model=List[StationSchema],
)
def stations_history(session: Session = Depends(get_database_session)):
    raise HTTPException(status_code=status.HTTP_501_NOT_IMPLEMENTED)


@router.get(
    "/station/history/{station_code}",
    name="station history",
    description="""Get history for a station""",
    response_model=StationSchema,
)
def station_history(station_code: str, session: Session = Depends(get_database_session)):
    raise HTTPException(status_code=status.HTTP_501_NOT_IMPLEMENTED)


@router.post(
    "/",
    name="station",
    # response_model=StationSchema,
    description="Create a station",
)
def station_create(
    session: Session = Depends(get_database_session),
    # station: StationSubmission = None,
):
    raise HTTPException(status_code=status.HTTP_501_NOT_IMPLEMENTED)


@router.put("/{station_id}", name="Station update")
def station_update(
    station_id: int,
    data: StationModification = {},
    session: Session = Depends(get_database_session),
) -> dict:
    station = session.query(Station).get(station_id)

    if not station:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Station not found")

    if data.modification == "approve":

        if station.approved is True:
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail="Station already approved",
            )

        station.approved = True
        station.approved_at = datetime.now()
        station.approved_by = "opennem.admin"

    if data.modification == "reject":
        station.approved = False

    session.add(station)
    session.commit()

    response = StationUpdateResponse(success=True, record=station)

    return response



# pylint: disable=no-member
"""
update facility_scada primary key to include is_forecast

Revision ID: 66e98f534a41
Revises: 07d27793b5e2
Create Date: 2021-01-25 17:43:00.755467

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "66e98f534a41"
down_revision = "07d27793b5e2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute("SET LOCAL synchronize_seqscans = off;")

    op.execute("ALTER TABLE facility_scada DROP CONSTRAINT facility_scada_pkey CASCADE")

    # ensure is_forecast has values
    op.execute("update facility_scada set is_forecast = False where is_forecast is null")

    op.create_primary_key(
        "facility_scada_pkey",
        "facility_scada",
        ["network_id", "trading_interval", "facility_code", "is_forecast"],
    )


def downgrade() -> None:
    op.execute("SET LOCAL synchronize_seqscans = off;")

    op.execute("ALTER TABLE facility_scada DROP CONSTRAINT facility_scada_pkey CASCADE")
    op.create_primary_key(
        "facility_scada_pkey",
        "facility_scada",
        ["network_id", "trading_interval", "facility_code"],
    )



# pylint: disable=no-member
"""
Update vw_region_flow_emissions

Revision ID: d13c12e02b4e
Revises: 4503ef797f3a
Create Date: 2021-01-28 01:00:02.168088

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "d13c12e02b4e"
down_revision = "4503ef797f3a"
branch_labels = None
depends_on = None


stmt_drop = "drop materialized view if exists mv_region_emissions;"

stmt = """
create materialized view mv_region_emissions as select
    f.trading_interval,
    f.network_region,
    case
        when sum(f.energy) > 0 then
            sum(f.emissions) / sum(f.energy)
        else 0
    end as emissions_per_kw
from mv_facility_all f
group by 1, f.network_region
order by 1 desc;
"""

stmt_drop2 = "drop view if exists vw_region_flow_emissions;"

stmt2 = """
create view vw_region_flow_emissions as select
    f.trading_interval at time zone 'AEST' as trading_interval,
    date_trunc('day', f.trading_interval) at time zone 'AEST' as ti_day_aest,
    date_trunc('month', f.trading_interval) at time zone 'AEST' as ti_month_aest,
    sum(f.energy) as flow_energy,
    f.network_region || '->' || f.interconnector_region_to as flow_region,
    f.network_region as flow_from,
    f.interconnector_region_to as flow_to,
    abs(sum(ef.emissions_per_kw) * sum(f.energy)) as flow_from_emissions,
    abs(sum(et.emissions_per_kw) * sum(f.energy)) as flow_to_emissions
from mv_facility_all f
left join mv_region_emissions ef on
    ef.trading_interval = f.trading_interval and
    ef.network_region = f.network_region
left join mv_region_emissions et on
    et.trading_interval = f.trading_interval and
    et.network_region = f.interconnector_region_to
where
    f.interconnector is True
group by 1, 2, 3, flow_region, 5, 6, f.interconnector_region_to
order by trading_interval desc, flow_from asc, flow_to asc;
"""


def upgrade() -> None:
    with op.get_context().autocommit_block():
        op.execute(stmt_drop)
        op.execute(stmt)
        op.execute(stmt_drop2)
        op.execute(stmt2)


def downgrade() -> None:
    op.execute(stmt_drop)
    op.execute(stmt_drop2)



from pydantic import BaseModel


class PropertyBaseModel(BaseModel):
    """
    Workaround for serializing properties with pydantic until
    https://github.com/samuelcolvin/pydantic/issues/935
    is solved
    """

    @classmethod
    def get_properties(cls):
        return [
            prop
            for prop in dir(cls)
            if isinstance(getattr(cls, prop), property) and prop not in ("__values__", "fields")
        ]

    def dict(self, *args, **kwargs) -> "DictStrAny":
        self.__dict__.update({prop: getattr(self, prop) for prop in self.get_properties()})

        return super().dict(*args, **kwargs)


class BaseConfig(PropertyBaseModel):
    class Config:
        orm_mode = True
        anystr_strip_whitespace = True

        arbitrary_types_allowed = True
        validate_assignment = True

        json_encoders = {
            # datetime: lambda v: v.isotime(),
            # Decimal: lambda v: float(v),
        }



"""
OpenNEM Monitor Facility First Seen

Scheduled process that checks for new DUIDs found in generation data
and alerts about them

"""
import logging
from datetime import datetime
from typing import List, Optional

from opennem.db import get_database_engine
from opennem.schema.core import BaseConfig

# from opennem.notifications.slack import slack_message

logger = logging.getLogger("opennem.monitors.facility_seen")


class FacilitySeen(BaseConfig):
    code: str
    network_id: str
    seen_first: Optional[datetime]
    seen_last: Optional[datetime]


def get_facility_first_seen() -> List[FacilitySeen]:
    """Run this and it'll check if there are new facilities in
    scada data and let you know which ones

    The query can be an expensive one so don't run to often.
    """

    engine = get_database_engine()

    __query = """
        select
            distinct fs.facility_code,
            fs.network_id
        from facility_scada fs
        where
            fs.facility_code not in (select distinct code from facility);
    """

    with engine.connect() as c:
        logger.debug(__query)
        row = list(c.execute(__query))

    records: List[FacilitySeen] = [FacilitySeen(code=r[0], network_id=r[1]) for r in row]

    return records


def facility_first_seen_check() -> None:
    return None


if __name__ == "__main__":
    facility_first_seen_check()



# pylint: disable=no-member
"""
Update facility interconnector status

Revision ID: 2ed53b064d85
Revises: ddb5d2a90896
Create Date: 2021-02-16 23:58:10.961598

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "2ed53b064d85"
down_revision = "ddb5d2a90896"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute("update facility set interconnector=False where interconnector is null")


def downgrade() -> None:
    pass



# pylint: disable=no-member
"""
Indexes for day truncs in NEM and WEM timezones

Revision ID: d7c22e0add08
Revises: df7308f58ee2
Create Date: 2020-12-02 02:00:14.521669

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d7c22e0add08"
down_revision = "df7308f58ee2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_index(
        "idx_facility_scada_trading_interval_perth_day",
        "facility_scada",
        [sa.text("date_trunc('day', trading_interval AT TIME ZONE 'AWST')")],
    )

    op.create_index(
        "idx_facility_scada_trading_interval_sydney_day",
        "facility_scada",
        [sa.text("date_trunc('day', trading_interval AT TIME ZONE 'AEST')")],
    )


def downgrade() -> None:
    op.drop_index("idx_facility_scada_trading_interval_perth_day")
    op.drop_index("idx_facility_scada_trading_interval_sydney_day")



import logging
from datetime import datetime
from itertools import groupby
from typing import Optional

from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


class AEMOMMSDudetailSummaryGrouper(object):
    def parse_date(self, date_string: str) -> Optional[datetime]:
        """

            `25/10/1998  12:00:00 am` => d

        """
        date_string_components = date_string.strip().split(" ")

        if len(date_string_components) < 2:
            raise Exception("Error parsing date: {}".format(date_string))

        date_part = date_string_components[0]

        dt = None

        try:
            dt = datetime.strptime(date_part, "%Y/%m/%d")
        except ValueError:
            raise Exception("Error parsing date: {}".format(date_string))

        # AEMO sets dates to this value when they mean None
        # if dt.year == 2999:
        # return None

        return dt

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        if not "tables" in item:
            logger.error(item)
            raise Exception("No tables passed to pipeline")

        tables = item["tables"]
        table = tables.pop()

        records = table["records"]

        records = [
            {
                "date_start": self.parse_date(i["START_DATE"]),
                "date_end": self.parse_date(i["END_DATE"]),
                "DUID": i["DUID"],
                "REGIONID": i["REGIONID"],
                "STATIONID": i["STATIONID"],
                "PARTICIPANTID": i["PARTICIPANTID"],
                "DISPATCHTYPE": i["DISPATCHTYPE"]
                # "DUID": i["DUID"],
                # **i,
            }
            for i in records
            # if i["DISPATCHTYPE"] == "GENERATOR"
        ]

        grouped_records = {}

        # First pass sorts facilities into stations
        for k, v in groupby(records, lambda x: (x["STATIONID"], x["DUID"])):
            key = k[0]
            duid = k[1]
            if not key in grouped_records:
                grouped_records[key] = {}
                grouped_records[key]["id"] = k[0]
                # grouped_records[key]["participant"] = v[0]["PARTICIPANTID"]
                grouped_records[key]["details"] = {}
                grouped_records[key]["facilities"] = []

            if not duid in grouped_records[key]["details"]:
                grouped_records[key]["details"][duid] = []

            grouped_records[key]["details"][duid] += list(v)

        # Second pass flatten the records and we should get start and end dates and a derived status
        for rec in grouped_records.keys():
            for facility_group, facility_group_records in grouped_records[rec][
                "details"
            ].items():

                date_end_min = min(
                    facility_group_records, key=lambda x: x["date_end"]
                )
                date_end_max = max(
                    facility_group_records, key=lambda x: x["date_end"]
                )
                date_start_min = min(
                    facility_group_records, key=lambda x: x["date_start"]
                )

                # print(date_end_min, date_start_min, date_end_max)

                grouped_rec = {
                    **date_end_max,
                    "date_start": date_start_min["date_start"],
                }

                if grouped_rec["date_end"].year == 2999:
                    grouped_rec["date_end"] = None

                grouped_records[rec]["facilities"].append(grouped_rec)

        grouped_records = [
            {"id": i, "facilities": v["facilities"]}
            for i, v in grouped_records.items()
        ]

        return grouped_records



from io import BytesIO

from PIL import Image


def img_to_buffer(img: Image) -> memoryview:
    """
    Convert image and save as JPEG

    @TODO jpeg settings in opennem.settings

    """
    buf = BytesIO()

    # convert all to RGP
    if img.mode in ("RGBA", "P", "LA"):
        img = img.convert("RGB")

    img.save(buf, format="JPEG")

    return buf.getbuffer()



# pylint: disable=no-member
"""
Material view for energy outputs

Revision ID: 674d602ecf59
Revises: d8bd4db7c53d
Create Date: 2020-12-14 18:40:18.740870

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "674d602ecf59"
down_revision = "d8bd4db7c53d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        """
    drop view if exists mv_facility_energy_hour cascade;

    CREATE OR REPLACE VIEW mv_facility_energy_hour WITH (timescaledb.continuous) AS
    select
        time_bucket('1 hour', fs.trading_interval) as trading_interval,
        fs.facility_code,
        energy_sum(fs.generated, '1 hour') * interval_size('1 hour', count(fs.generated)) as energy
    from facility_scada fs
    group by
        1,
        fs.facility_code;
    """
    )


def downgrade() -> None:
    op.execute(
        """
    drop view if exists mv_facility_energy_hour cascade;
    """
    )



"""
OpenNEM Excel File Utils


"""


from typing import Any

import xlrd
from openpyxl.workbook.workbook import Workbook


def convert_to_xlxs(content: Any) -> Workbook:
    """Convert old workbook formats xls into xlxs"""
    xlsBook = xlrd.open_workbook(file_contents=content)
    workbook = Workbook()

    for i in range(0, xlsBook.nsheets):
        xlsSheet = xlsBook.sheet_by_index(i)
        sheet = workbook.active if i == 0 else workbook.create_sheet()
        sheet.title = xlsSheet.name

        for row in range(0, xlsSheet.nrows):
            for col in range(0, xlsSheet.ncols):
                sheet.cell(
                    row=row + 1, column=col + 1
                ).value = xlsSheet.cell_value(row, col)

    return workbook



import logging
import sys

import requests
from validators import ValidationFailure
from validators.url import url as valid_url

from opennem.settings import settings
from opennem.settings.scrapy import USER_AGENT

logger = logging.getLogger(__name__)

REQ_HEADERS = {"User-Agent": USER_AGENT, "Content-type": "application/json"}


def slack_message(msg: str) -> bool:
    """
    Post a slack message to the watchdog channel

    """
    if not settings.slack_notifications:
        return False

    if not settings.slack_hook_url:
        logger.error("No slack notification endpoint configured")
        return False

    if isinstance(valid_url(settings.slack_hook_url), ValidationFailure):
        logger.error("No slack notification endpoint configured bad url")
        return False

    alert_url = settings.slack_hook_url

    resp = requests.post(alert_url, json={"text": msg})

    if resp.status_code != 200:
        logger.error("Error sending slack message")
        return False

    return True


if __name__ == "__main__":
    if len(sys.argv) < 2:
        sys.exit(1)

    msg_args = sys.argv[1:]
    slack_msg = " ".join([str(i) for i in msg_args])
    slack_message(slack_msg)



import logging
from typing import List

import aioredis
from fastapi import Depends, FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse
from fastapi_cache import FastAPICache
from fastapi_cache.backends.redis import RedisBackend
from prometheus_fastapi_instrumentator import Instrumentator
from sqlalchemy.orm import Session
from starlette import status

from opennem.api.admin.router import router as admin_router
from opennem.api.export.router import router as export_router
from opennem.api.facility.router import router as facility_router
from opennem.api.geo.router import router as geo_router
from opennem.api.locations import router as locations_router
from opennem.api.revision.router import router as revisions_router
from opennem.api.station.router import router as station_router
from opennem.api.stats.router import router as stats_router
from opennem.api.tasks.router import router as tasks_router
from opennem.api.weather.router import router as weather_router
from opennem.core.time import INTERVALS, PERIODS
from opennem.core.units import UNITS
from opennem.db import database, get_database_session
from opennem.db.models.opennem import FuelTech, Network, NetworkRegion
from opennem.schema.network import NetworkRegionSchema, NetworkSchema
from opennem.schema.opennem import FueltechSchema
from opennem.schema.time import TimeInterval, TimePeriod
from opennem.schema.units import UnitDefinition
from opennem.settings import settings
from opennem.utils.http_cache import PydanticCoder
from opennem.utils.version import get_version

logger = logging.getLogger(__name__)


app = FastAPI(title="OpenNEM", debug=settings.debug, version=get_version())
Instrumentator().instrument(app).expose(app)

try:
    from fastapi.staticfiles import StaticFiles

    app.mount(
        "/static",
        StaticFiles(directory=settings.static_folder_path),
        name="static",
    )
except Exception as e:
    logger.error("Error initializing static hosting: {}".format(e))


app.include_router(stats_router, tags=["Stats"], prefix="/stats")
app.include_router(locations_router, tags=["Locations"], prefix="/locations")
app.include_router(geo_router, tags=["Geo"], prefix="/geo")
app.include_router(station_router, tags=["Stations"], prefix="/station")
app.include_router(facility_router, tags=["Facilities"], prefix="/facility")
app.include_router(revisions_router, tags=["Revisions"], prefix="/revision")
app.include_router(weather_router, tags=["Weather"], prefix="/weather")
app.include_router(admin_router, tags=["Admin"], prefix="/admin")
app.include_router(export_router, tags=["Export"], prefix="/export")
app.include_router(tasks_router, tags=["Tasks"], prefix="/tasks")


origins = [
    "https://opennem.org.au",
    "https://dev.opennem.org.au",
    "https://admin.opennem.org.au",
    "https://admin.opennem.test",
    "http://localhost:8001",
    "http://127.0.0.1:8001",
    "http://127.0.0.1:8002",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.on_event("startup")
async def startup() -> None:
    logger.debug("In startup")
    # await database.connect()

    redis = await aioredis.create_redis_pool(settings.cache_url, encoding="utf-8")
    FastAPICache.init(RedisBackend(redis), prefix="api-cache", coder=PydanticCoder)


# @app.on_event("shutdown")
async def shutdown() -> None:
    logger.debug("In shutdown")
    await database.disconnect()


@app.get("/robots.txt", response_class=FileResponse)
def robots_txt() -> FileResponse:
    if not settings.debug:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return FileResponse(settings.static_folder_path + "/robots.txt")


@app.get("/networks", response_model=List[NetworkSchema])
def networks(
    session: Session = Depends(get_database_session),
) -> List[NetworkSchema]:
    networks = session.query(Network).all()

    if not networks:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    response = [NetworkSchema.parse_obj(i) for i in networks]

    return response


@app.get("/networks/regions", response_model=List[NetworkRegionSchema])
def network_regions(
    session: Session = Depends(get_database_session),
) -> List[NetworkRegionSchema]:
    networks = session.query(NetworkRegion).all()

    if not networks:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    response = [NetworkRegionSchema.parse_obj(i) for i in networks]

    return response


@app.get("/fueltechs", response_model=List[FueltechSchema])
def fueltechs(
    session: Session = Depends(get_database_session),
) -> List[FueltechSchema]:
    fueltechs = session.query(FuelTech).all()

    if not fueltechs:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    response = [FueltechSchema.parse_obj(i) for i in fueltechs]

    return response


@app.get("/intervals", response_model=List[TimeInterval])
def intervals() -> List[TimeInterval]:
    return INTERVALS


@app.get("/periods", response_model=List[TimePeriod])
def periods() -> List[TimePeriod]:
    return PERIODS


@app.get("/units", response_model=List[UnitDefinition])
def units() -> List[UnitDefinition]:
    return UNITS



import logging
from typing import List, Optional

from opennem.api.export.queries import (
    country_stats_query,
    energy_network_flow_query,
    energy_network_fueltech_query,
    energy_network_interconnector_emissions_query,
    interconnector_power_flow,
    network_demand_query,
    power_network_fueltech_query,
    power_network_rooftop_query,
    price_network_query,
    weather_observation_query,
)
from opennem.api.stats.controllers import stats_factory
from opennem.api.stats.schema import DataQueryResult, OpennemDataSet, RegionFlowEmissionsResult
from opennem.api.time import human_to_interval, human_to_period
from opennem.core.flows import net_flows_emissions
from opennem.core.units import get_unit
from opennem.db import get_database_engine
from opennem.schema.dates import TimeSeries
from opennem.schema.network import NetworkNEM, NetworkSchema
from opennem.schema.stats import StatTypes
from opennem.schema.time import TimePeriod

logger = logging.getLogger(__name__)


def weather_daily(
    time_series: TimeSeries,
    station_code: str,
    unit_name: str = "temperature_mean",
    include_min_max: bool = True,
    network_region: Optional[str] = None,
) -> Optional[OpennemDataSet]:
    engine = get_database_engine()
    units = get_unit(unit_name)

    query = weather_observation_query(
        time_series=time_series,
        station_codes=[station_code],
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    temp_avg = [
        DataQueryResult(interval=i[0], group_by=i[1], result=i[2] if len(i) > 1 else None)
        for i in row
    ]

    temp_min = [
        DataQueryResult(interval=i[0], group_by=i[1], result=i[3] if len(i) > 1 else None)
        for i in row
    ]

    temp_max = [
        DataQueryResult(interval=i[0], group_by=i[1], result=i[4] if len(i) > 1 else None)
        for i in row
    ]

    if len(temp_avg) < 1:
        logger.info("No results from query: {}".format(query))
        return None

    stats = stats_factory(
        stats=temp_avg,
        units=units,
        network=time_series.network,
        interval=time_series.interval,
        region=network_region,
        code="bom",
        group_field="temperature",
    )

    if not stats:
        return None

    if include_min_max:
        stats_min = stats_factory(
            stats=temp_min,
            units=get_unit("temperature_min"),
            network=time_series.network,
            interval=time_series.interval,
            region=network_region,
            code="bom",
            group_field="temperature",
        )

        stats_max = stats_factory(
            stats=temp_max,
            units=get_unit("temperature_max"),
            network=time_series.network,
            interval=time_series.interval,
            region=network_region,
            code="bom",
            group_field="temperature",
        )

        stats.append_set(stats_min)
        stats.append_set(stats_max)

    return stats


def gov_stats_cpi() -> Optional[OpennemDataSet]:
    engine = get_database_engine()

    query = country_stats_query(StatTypes.CPI)

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    stats = [
        DataQueryResult(interval=i[0], result=i[1], group_by=i[2] if len(i) > 1 else None)
        for i in row
    ]

    if len(stats) < 1:
        raise Exception("No results from query: {}".format(query))

    result = stats_factory(
        stats,
        code="au.cpi",
        network=NetworkNEM,
        interval=human_to_interval("1Q"),
        period=human_to_period("all"),
        units=get_unit("cpi"),
        group_field="gov",
    )

    return result


def power_flows_week(
    time_series: TimeSeries,
    network_region_code: str,
) -> Optional[OpennemDataSet]:
    engine = get_database_engine()

    query = interconnector_power_flow(time_series=time_series, network_region=network_region_code)

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    if len(row) < 1:
        raise Exception("No results from query: {}".format(query))

    imports = [
        DataQueryResult(interval=i[0], result=i[2], group_by="imports" if len(i) > 1 else None)
        for i in row
    ]

    exports = [
        DataQueryResult(interval=i[0], result=i[3], group_by="exports" if len(i) > 1 else None)
        for i in row
    ]

    result = stats_factory(
        imports,
        # code=network_region_code or network.code,
        network=time_series.network,
        period=human_to_period("7d"),
        interval=human_to_interval("5m"),
        units=get_unit("power"),
        region=network_region_code,
        fueltech_group=True,
    )

    if not result:
        raise Exception("No results")

    result_exports = stats_factory(
        exports,
        # code=network_region_code or network.code,
        network=time_series.network,
        period=human_to_period("7d"),
        interval=human_to_interval("5m"),
        units=get_unit("power"),
        region=network_region_code,
        fueltech_group=True,
    )

    result.append_set(result_exports)

    return result


def demand_week(
    time_series: TimeSeries,
    network_region_code: Optional[str],
    networks_query: Optional[List[NetworkSchema]] = None,
) -> Optional[OpennemDataSet]:
    engine = get_database_engine()

    query = network_demand_query(
        time_series=time_series,
        network_region=network_region_code,
        networks_query=networks_query,
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    if len(row) < 1:
        raise Exception("No results from query: {}".format(query))

    demand = [
        DataQueryResult(interval=i[0], result=i[2], group_by="demand" if len(i) > 1 else None)
        for i in row
    ]

    result = stats_factory(
        demand,
        # code=network_region_code or network.code,
        network=time_series.network,
        period=human_to_period("7d"),
        interval=human_to_interval("5m"),
        units=get_unit("demand"),
        region=network_region_code,
    )

    if not result:
        raise Exception("No results")

    return result


def power_week(
    time_series: TimeSeries,
    network_region_code: str = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> Optional[OpennemDataSet]:
    engine = get_database_engine()

    query = power_network_fueltech_query(
        time_series=time_series,
        networks_query=networks_query,
        network_region=network_region_code,
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    stats = [
        DataQueryResult(interval=i[0], result=i[2], group_by=i[1] if len(i) > 1 else None)
        for i in row
    ]

    if len(stats) < 1:
        raise Exception("No results from query: {}".format(query))

    result = stats_factory(
        stats,
        # code=network_region_code or network.code,
        network=time_series.network,
        interval=time_series.interval,
        period=time_series.period,
        units=get_unit("power"),
        region=network_region_code,
        fueltech_group=True,
    )

    if not result:
        raise Exception("No results")

    # price

    time_series_price = time_series.copy()
    time_series_price.interval = human_to_interval("30m")

    query = price_network_query(
        time_series=time_series_price,
        networks_query=networks_query,
        network_region=network_region_code,
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    stats_price = [
        DataQueryResult(interval=i[0], result=i[2], group_by=i[1] if len(i) > 1 else None)
        for i in row
    ]

    stats_market_value = stats_factory(
        stats=stats_price,
        code=network_region_code or time_series.network.code.lower(),
        units=get_unit("price_energy_mega"),
        network=time_series.network,
        interval=human_to_interval("30m"),
        region=network_region_code,
        period=time_series.period,
    )

    result.append_set(stats_market_value)

    # rooftop solar

    time_series_rooftop = time_series.copy()
    time_series_rooftop.interval = human_to_interval("30m")

    query = power_network_rooftop_query(
        time_series=time_series_rooftop,
        networks_query=networks_query,
        network_region=network_region_code,
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    rooftop_power = [
        DataQueryResult(interval=i[0], result=i[2], group_by=i[1] if len(i) > 1 else None)
        for i in row
    ]

    rooftop = stats_factory(
        rooftop_power,
        # code=network_region_code or network.code,
        network=time_series.network,
        interval=human_to_interval("30m"),
        period=time_series.period,
        units=get_unit("power"),
        region=network_region_code,
        fueltech_group=True,
    )

    # rooftop forecast

    time_series_rooftop_forecast = time_series.copy()
    time_series_rooftop_forecast.interval = human_to_interval("30m")
    time_series_rooftop_forecast.forecast = True

    query = power_network_rooftop_query(
        time_series=time_series_rooftop,
        networks_query=networks_query,
        network_region=network_region_code,
        forecast=True,
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    rooftop_forecast_power = [
        DataQueryResult(interval=i[0], result=i[2], group_by=i[1] if len(i) > 1 else None)
        for i in row
    ]

    rooftop_forecast = stats_factory(
        rooftop_forecast_power,
        # code=network_region_code or network.code,
        network=time_series.network,
        interval=human_to_interval("30m"),
        period=time_series.period,
        units=get_unit("power"),
        region=network_region_code,
        fueltech_group=True,
    )

    if rooftop and rooftop_forecast:
        if (
            hasattr(rooftop, "data")
            and len(rooftop.data) > 0
            and rooftop_forecast.data
            and len(rooftop_forecast.data) > 0
        ):
            rooftop.data[0].forecast = rooftop_forecast.data[0].history

    result.append_set(rooftop)

    return result


def energy_fueltech_daily(
    time_series: TimeSeries,
    network_region_code: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> Optional[OpennemDataSet]:
    engine = get_database_engine()
    units = get_unit("energy_giga")

    query = energy_network_fueltech_query(
        time_series=time_series,
        network_region=network_region_code,
        networks_query=networks_query,
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    results_energy = [
        DataQueryResult(interval=i[0], group_by=i[1], result=i[2] if len(i) > 1 else None)
        for i in row
    ]

    results_market_value = [
        DataQueryResult(interval=i[0], group_by=i[1], result=i[3] if len(i) > 1 else None)
        for i in row
    ]

    results_emissions = [
        DataQueryResult(interval=i[0], group_by=i[1], result=i[4] if len(i) > 1 else None)
        for i in row
    ]

    if len(results_energy) < 1:
        raise Exception("No results from query: {}".format(query))

    stats = stats_factory(
        stats=results_energy,
        units=units,
        network=time_series.network,
        fueltech_group=True,
        interval=time_series.interval,
        region=network_region_code,
        period=time_series.period,
        localize=False,
        # code=network.code.lower(),
    )

    if not stats:
        return None

    stats_market_value = stats_factory(
        stats=results_market_value,
        units=get_unit("market_value"),
        network=time_series.network,
        fueltech_group=True,
        interval=time_series.interval,
        region=network_region_code,
        period=time_series.period,
        code=time_series.network.code.lower(),
        localize=False,
    )

    stats.append_set(stats_market_value)

    stats_emissions = stats_factory(
        stats=results_emissions,
        units=get_unit("emissions"),
        network=time_series.network,
        fueltech_group=True,
        interval=time_series.interval,
        region=network_region_code,
        period=time_series.period,
        code=time_series.network.code.lower(),
        localize=False,
    )

    stats.append_set(stats_emissions)

    return stats


def energy_interconnector_region_daily(
    time_series: TimeSeries,
    network_region_code: str,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> Optional[OpennemDataSet]:
    engine = get_database_engine()
    period: TimePeriod = human_to_period("1Y")
    units = get_unit("energy_giga")

    query = energy_network_flow_query(
        time_series=time_series,
        network_region=network_region_code,
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    if len(row) < 1:
        return None

    imports = [DataQueryResult(interval=i[0], group_by="imports", result=i[1]) for i in row]

    exports = [DataQueryResult(interval=i[0], group_by="exports", result=i[2]) for i in row]

    result = stats_factory(
        imports,
        # code=network_region_code or network.code,
        network=time_series.network,
        period=period,
        interval=time_series.interval,
        units=units,
        region=network_region_code,
        fueltech_group=True,
        # localize=False,
    )

    # Bail early on no interconnector
    # don't error
    if not result:
        logger.warn("No interconnector energy result")
        return result

    result_exports = stats_factory(
        exports,
        # code=network_region_code or network.code,
        network=time_series.network,
        period=period,
        interval=time_series.interval,
        units=units,
        region=network_region_code,
        fueltech_group=True,
        # localize=False,
    )

    result.append_set(result_exports)

    return result


def energy_interconnector_emissions_region_daily(
    time_series: TimeSeries,
    network_region_code: str,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> Optional[OpennemDataSet]:
    engine = get_database_engine()
    period: TimePeriod = human_to_period("1Y")
    units = get_unit("emissions")

    query = energy_network_interconnector_emissions_query(
        time_series=time_series,
        network_region=network_region_code,
        networks_query=networks_query,
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    if len(row) < 1:
        return None

    stats = [
        RegionFlowEmissionsResult(
            interval=i[0],
            flow_from=i[1],
            flow_to=i[2],
            energy=i[3],
            flow_from_emissions=i[4],
            flow_to_emissions=i[5],
        )
        for i in row
    ]

    stats_grouped = net_flows_emissions(network_region_code, stats, time_series.interval)

    imports = stats_grouped["imports"]
    exports = stats_grouped["exports"]

    result = stats_factory(
        imports,
        # code=network_region_code or network.code,
        network=time_series.network,
        period=period,
        interval=time_series.interval,
        units=units,
        region=network_region_code,
        fueltech_group=True,
        localize=False,
    )

    # Bail early on no interconnector
    # don't error
    if not result:
        return result

    result_exports = stats_factory(
        exports,
        # code=network_region_code or network.code,
        network=time_series.network,
        period=period,
        interval=time_series.interval,
        units=units,
        region=network_region_code,
        fueltech_group=True,
        localize=False,
    )

    result.append_set(result_exports)

    return result



"""
Utilities to work on series

"""

from datetime import datetime
from math import isclose
from operator import add
from typing import Dict, List, Tuple

from opennem.utils.dates import strip_timezone


def add_series(*args) -> List:
    return list(map(add, *args))


def series_diff(s1, s2):
    s2d = {i[0]: i[1] for i in s2}
    s1d = {i[0]: i[1] for i in s1}
    d = dict()
    for k in s1d.keys():
        if k in s2d:
            if isinstance(s2d[k], float) and isinstance(s1d[k], float):
                d[k] = s2d[k] - s1d[k]
            elif isinstance(s1d[k], float):
                d[k] = -1 * s1d[k]
            elif isinstance(s2d[k], float):
                d[k] = s2d[k]

    return list(d.keys()), list(d.values())


def series_diff_percentage(s1, s2):
    s2d = {i[0]: i[1] for i in s2}
    s1d = {i[0]: i[1] for i in s1}
    d = dict()
    for k in s1d.keys():
        if k in s2d:
            if isinstance(s2d[k], float) and isinstance(s1d[k], float):
                if s2d[k] == s1d[k]:
                    d[k] = 0

                if s1d[k] == 0 and s2d[k] > 0:
                    d[k] = 100

                if s1d[k] != 0:
                    try:
                        d[k] = (abs(s1d[k]) - abs(s2d[k])) / abs(s1d[k]) * 100 * -1
                    except ZeroDivisionError:
                        pass

            elif isinstance(s1d[k], float):
                d[k] = 100
            elif isinstance(s2d[k], float):
                d[k] = 100

    return list(d.keys()), list(d.values())


def are_approx_equal(actual: float, desired: float) -> bool:
    if actual == desired:
        return True

    if not desired or not actual:
        return True

    return isclose(actual, desired, abs_tol=0.5, rel_tol=0.0001)


def series_are_equal(
    s1: List[Tuple[datetime, float]], s2: List[Tuple[datetime, float]], full_equality: bool = False
) -> Dict:
    s2d = {strip_timezone(i[0]): i[1] for i in s2}
    s1d = {strip_timezone(i[0]): i[1] for i in s1}
    d = dict()

    for k in s1d.keys():
        if k in s2d:
            if full_equality:
                d[k] = s1d[k] == s2d[k]
            else:
                d[k] = are_approx_equal(s1d[k], s2d[k])

    return d


def series_joined(
    s1: List[Tuple[datetime, float]], s2: List[Tuple[datetime, float]], full_equality: bool = False
) -> Dict:
    s2d = {strip_timezone(i[0]): i[1] for i in s2}
    s1d = {strip_timezone(i[0]): i[1] for i in s1}
    d = dict()

    for k in s1d.keys():
        if k in s2d:
            d[k.isoformat()] = {"v2": s1d[k], "v3": s2d[k]}

    return d


def series_not_close(
    s1: List[Tuple[datetime, float]],
    s2: List[Tuple[datetime, float]],
    full_equality: bool = False,
) -> Dict:
    s2d = {strip_timezone(i[0]): i[1] for i in s2}
    s1d = {strip_timezone(i[0]): i[1] for i in s1}
    d = dict()

    for k in s1d.keys():
        if k in s2d:
            if full_equality:
                if s1d[k] != s2d[k]:
                    if s2d[k] and s1d[k]:
                        d[k.isoformat()] = {"v2": s1d[k], "v3": s2d[k]}
            else:
                if not are_approx_equal(s1d[k], s2d[k]):
                    if s2d[k] and s1d[k]:
                        d[k.isoformat()] = {"v2": s1d[k], "v3": s2d[k]}

    return d


def series_trim_to_date(
    s1: List[Tuple[datetime, float]],
    s2: List[Tuple[datetime, float]],
) -> Tuple[List, List]:
    s2d = {strip_timezone(i[0]): i[1] for i in s2}
    s1d = {strip_timezone(i[0]): i[1] for i in s1}
    d1 = []
    d2 = []

    for k in s1d.keys():
        if k in s2d:
            if s1d[k] and s2d[k]:
                d1.append(s1d[k])
                d2.append(s2d[k])

    return d1, d2



# pylint: disable=no-member
"""
Update mv_facility_energy_hour

Revision ID: 60ff840c16c5
Revises: 31ea8ecdfa44
Create Date: 2021-02-02 02:05:14.944479

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "60ff840c16c5"
down_revision = "31ea8ecdfa44"
branch_labels = None
depends_on = None

stmt_drop = "drop materialized view if exists mv_facility_energy_hour cascade;"


stmt = """
CREATE MATERIALIZED VIEW mv_facility_energy_hour WITH (timescaledb.continuous) AS
select
    time_bucket('1 hour', fs.trading_interval) as trading_interval,
    fs.facility_code,
    fs.network_id,
    (case
        when count(fs.eoi_quantity) > 0 then sum(fs.eoi_quantity)
        else energy_sum(fs.generated, '1 hour') * interval_size('1 hour', count(fs.generated))
    end) as energy
from facility_scada fs
where fs.is_forecast is False
group by
    1, 2, 3;
"""


def upgrade() -> None:
    with op.get_context().autocommit_block():
        op.execute(stmt_drop)
        op.execute(stmt)


def downgrade() -> None:
    op.execute(stmt_drop)



import sys

# Check minimum required Python version
if sys.version_info < (3, 7):
    print("OpenNEM %s requires Python 3.7 or greater")
    sys.exit(1)


# setup sentry
from opennem.settings import settings
from opennem.utils.sentry import setup_sentry

if settings.sentry_enabled:
    setup_sentry()


import warnings

# Ignore noisy twisted deprecation warnings
warnings.filterwarnings(
    "ignore", category=DeprecationWarning, module="twisted"
)



# pylint: disable=no-member
"""
update bom_observation add temp_min and temp_max

Revision ID: c203c9207aa3
Revises: f048f1d8922c
Create Date: 2021-01-28 20:10:05.686924

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c203c9207aa3"
down_revision = "f048f1d8922c"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column("bom_observation", sa.Column("temp_max", sa.Numeric(), nullable=True))
    op.add_column("bom_observation", sa.Column("temp_min", sa.Numeric(), nullable=True))


def downgrade() -> None:
    op.drop_column("bom_observation", "temp_min")
    op.drop_column("bom_observation", "temp_max")



from pathlib import Path
from typing import List

from opennem.core.parsers.aemo import AEMOTableSet, parse_aemo_csv

MMS_DATA_DIR = "~/Projects/Opennem/data/mms/"
MMS_DATA_PREFIX = "MMSDM_"
MMS_APPEND_DATA_DIR_PATH = "MMSDM_Historical_Data_SQLLoader/DATA"
MMS_HISTORIC_DATA_DIR_PATH = "MMSDM_Historical_Data_SQLLoader/DATA"


def get_mms_sets(data_dir: str = MMS_DATA_DIR) -> List[Path]:
    """
    Finds MMS data sets in a directory and returns
    all paths sorted in chronological order

    """

    data_path = Path(data_dir).expanduser().resolve()

    if not data_path.is_dir():
        raise Exception("Not a directory: {}".format(data_dir))

    mms_files = [
        d
        for d in data_path.iterdir()
        if d.is_dir()
        and d.name.startswith(MMS_DATA_PREFIX)
        and (d / MMS_HISTORIC_DATA_DIR_PATH).is_dir()
    ]

    mms_files = sorted(mms_files)

    return mms_files


def import_aemo_mms():
    mms_sets = get_mms_sets()

    for x in mms_sets:
        print(x)


if __name__ == "__main__":
    import_aemo_mms()



# pylint: disable=no-member
"""
create idx_mv_region_emissions_value

Revision ID: a6261453b8e7
Revises: c203c9207aa3
Create Date: 2021-01-29 02:12:58.397189

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "a6261453b8e7"
down_revision = "c203c9207aa3"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        "CREATE UNIQUE INDEX IF NOT EXISTS idx_mv_region_emissions_unique ON mv_region_emissions (trading_interval, network_region);"
    )


def downgrade() -> None:
    op.execute("drop index if exists idx_mv_region_emissions_unique")



import csv
import logging
from datetime import datetime, timedelta

from sqlalchemy.dialects.postgresql import insert

from opennem.core.normalizers import normalize_duid
from opennem.db import SessionLocal, get_database_engine
from opennem.db.models.opennem import FacilityScada
from opennem.pipelines.nem.opennem import unit_scada_generate_facility_scada
from opennem.schema.network import NetworkWEM
from opennem.utils.dates import parse_date
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


class WemStoreFacilityScada(object):
    @check_spider_pipeline
    def process_item(self, item, spider=None):
        if "content" not in item:
            logger.error("No item content slipping store facility scada")
            return item

        csvreader = csv.DictReader(item["content"].split("\n"))

        item["table_schema"] = FacilityScada
        item["update_fields"] = ["generated", "eoi_quantity"]
        item["records"] = unit_scada_generate_facility_scada(
            csvreader,
            spider,
            interval_field="Trading Interval",
            facility_code_field="Facility Code",
            power_field="EOI Quantity (MW)",
            energy_field="Energy Generated (MWh)",
            network=NetworkWEM,
        )
        item["content"] = None

        return item


class WemStoreFacilityIntervals(object):
    @check_spider_pipeline
    def process_item(self, item, spider=None):
        if "content" not in item:
            logger.error("No item content slipping store facility scada")
            return item

        csvreader = csv.DictReader(item["content"].split("\n"))

        item["table_schema"] = FacilityScada
        item["update_fields"] = ["generated", "eoi_quantity"]
        item["records"] = unit_scada_generate_facility_scada(
            csvreader,
            spider,
            interval_field="PERIOD",
            facility_code_field="FACILITY_CODE",
            power_field="ACTUAL_MW",
            network=NetworkWEM,
        )
        item["content"] = None

        return item


class WemStoreLiveFacilityScada(object):
    """
    Store live facility scada data.

    @NOTE no longer used

    """

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        session = SessionLocal()
        engine = get_database_engine()

        csvreader = csv.DictReader(item["content"].split("\n"))

        records_to_store = []
        last_asat = None

        for row in csvreader:

            # @TODO MAX_GEN_CAPACITY
            # facility_capacity = row["MAX_GEN_CAPACITY"]

            if row["AS_AT"] != "":
                last_asat = parse_date(row["AS_AT"], network=NetworkWEM, dayfirst=False)

            if not last_asat or type(last_asat) is not datetime:
                logger.error("Invalid row or no datetime")
                continue

            # We need to pivot the table since columns are time intervals
            for i in range(1, 48):

                column = f"I{i:02}"

                if column not in row:
                    logger.error("Do not have data for interval {}".format(column))
                    continue

                if i > 0:
                    interval = last_asat - timedelta(minutes=(i - 1) * 30)
                else:
                    interval = last_asat

                facility_code = normalize_duid(row["FACILITY_CODE"])

                val = None

                try:
                    val = float(row[column]) / 2 or None
                except ValueError:
                    pass

                records_to_store.append(
                    {
                        "created_by": spider.name,
                        "network_id": "WEM",
                        "trading_interval": interval,
                        "facility_code": facility_code,
                        "eoi_quantity": val,
                    }
                )

        stmt = insert(FacilityScada).values(records_to_store)
        stmt.bind = engine
        stmt = stmt.on_conflict_do_update(
            index_elements=["trading_interval", "network_id", "facility_code", "is_forecast"],
            set_={
                # "updated_by": stmt.excluded.created_by,
                "eoi_quantity": stmt.excluded.eoi_quantity,
            },
        )

        try:
            session.execute(stmt)
            session.commit()
        except Exception as e:
            logger.error("Error inserting records")
            logger.error(e)
        finally:
            session.close()

        return len(records_to_store)



"""
Scrapy settings for nemweb project

For simplicity, this file contains only settings considered important or
commonly used. You can find more settings consulting the documentation:

https://docs.scrapy.org/en/latest/topics/settings.html
"""

BOT_NAME = "opennem"
SPIDER_MODULES = ["opennem.spiders"]
NEWSPIDER_MODULE = "opennem.spiders"

LOG_LEVEL = "ERROR"
LOG_STDOUT = False
# L

USER_AGENT = "opennem/v3.0.0 (+https://opennem.org.au)"
ROBOTSTXT_OBEY = False
CONCURRENT_REQUESTS = 8
COOKIES_ENABLED = True
TELNETCONSOLE_ENABLED = False

DEFAULT_REQUEST_HEADERS = {
    "Accept": "*/*",
    "Accept-Language": "en",
}

AUTOTHROTTLE_ENABLED = True
AUTOTHROTTLE_START_DELAY = 2
AUTOTHROTTLE_MAX_DELAY = 60
AUTOTHROTTLE_TARGET_CONCURRENCY = 2.0

DNSCACHE_ENABLED = True
DNS_TIMEOUT = 10

# SPIDER_MODULES = ["opennem.spiders"]
SPIDER_LOADER_WARN_ONLY = True

# Pipline docs:
# 1xx series - download handlers
# 2xx series - unpackers
# 3xx series -
# 4xx series - database processes
ITEM_PIPELINES = {
    # Generic Pipelines
    "opennem.pipelines.files.LinkExtract": 101,
    "opennem.pipelines.nem.UnzipSingleFilePipeline": 302,
    "opennem.pipelines.nem.ReadStringHandle": 310,
    "opennem.pipelines.nem.ExtractCSV": 320,
    "opennem.pipelines.nem.TableRecordSplitter": 330,
    # AEMO XLS Parsers
    "opennem.pipelines.aemo.registration_exemption.RegistrationExemptionGrouperPipeline": 350,
    "opennem.pipelines.aemo.general_information.GeneralInformationGrouperPipeline": 351,
    "opennem.pipelines.aemo.registration_exemption.RegistrationExemptionStorePipeline": 355,
    "opennem.pipelines.aemo.general_information.GeneralInformationStoragePipeline": 356,
    # Opennem storers
    "opennem.pipelines.nem.DatabaseStore": 400,
    "opennem.pipelines.nem.opennem.NemwebUnitScadaOpenNEMStorePipeline": 405,
    "opennem.pipelines.wem.facilities.WemStoreFacility": 410,
    "opennem.pipelines.wem.facility_scada.WemStoreFacilityScada": 411,
    "opennem.pipelines.wem.balancing_summary.WemStoreBalancingSummary": 412,
    "opennem.pipelines.wem.pulse.WemStorePulse": 414,
    "opennem.pipelines.wem.facility_scada.WemStoreLiveFacilityScada": 415,
    "opennem.pipelines.wem.facility_scada.WemStoreFacilityIntervals": 416,
    "opennem.pipelines.wem.facilities.WemStoreLiveFacilities": 417,
    "opennem.pipelines.wem.participant.WemStoreParticipant": 418,
    "opennem.pipelines.wem.participant.WemStoreLiveParticipant": 419,
    "opennem.pipelines.nem.summary.NemwebSummaryPipeline": 420,
    # APVI
    "opennem.pipelines.apvi.data.APVIStoreData": 450,
    # NPI
    "opennem.pipelines.npi.facilities.NPIStoreFacility": 455,
    # BOM Storers
    "opennem.pipelines.bom.StoreBomObservation": 470,
    # MMS Parsers and Storers
    "opennem.pipelines.nem.mms_dudetailsummary.AEMOMMSDudetailSummaryGrouper": 504,
    "opennem.pipelines.nem.mms.NemStoreMMSStations": 551,
    "opennem.pipelines.nem.mms.NemStoreMMSStationStatus": 552,
    "opennem.pipelines.nem.mms.NemStoreMMSDudetail": 553,
    "opennem.pipelines.nem.mms.NemStoreMMSDudetailSummary": 554,
    "opennem.pipelines.nem.mms.NemStoreMMSParticipant": 555,
    # DB bulk Inserterers
    "opennem.pipelines.csv.RecordsToCSVPipeline": 610,
    "opennem.pipelines.bulk_insert.BulkInsertPipeline": 620,
}

# @TODO if DEBUG
HTTPCACHE_ENABLED = False
# HTTPCACHE_DIR = ".scrapy"

RETRY_ENABLED = True
RETRY_TIMES = 9
RETRY_HTTP_CODES = [400, 403, 500, 502, 503, 504, 522, 524, 408, 429]



# pylint: disable=no-member
"""
Upate price field

Revision ID: ddb5d2a90896
Revises: 57c6030dcb87
Create Date: 2021-02-10 01:50:27.287152

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ddb5d2a90896"
down_revision = "57c6030dcb87"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        "balancing_summary",
        sa.Column("price_dispatch", sa.Numeric(), nullable=True),
    )


def downgrade() -> None:
    op.drop_column("balancing_summary", "price_dispatch")



# pylint: disable=no-member
"""
Update indexes on mvs

Revision ID: 405b2ea7b351
Revises: 12c8cbba29f0
Create Date: 2021-01-19 17:17:22.558587

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "405b2ea7b351"
down_revision = "12c8cbba29f0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute("drop index if exists idx_mv_facility_all_trading_interval_day_aest;")

    op.execute(
        "CREATE UNIQUE INDEX IF NOT EXISTS idx_mv_facility_unique ON mv_facility_all (trading_interval, network_id, code);"
    )


def downgrade() -> None:
    pass



# pylint: disable=no-member
"""
Fix APVI scada records to have network APVI

Revision ID: 467a3223b204
Revises: bfb98983b92c
Create Date: 2020-12-07 16:37:00.632028

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "467a3223b204"
down_revision = "bfb98983b92c"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        "delete from facility_scada where network_id='NEM' and facility_code like 'ROOFTOP_APVI_%'"
    )


def downgrade() -> None:
    pass



# pylint: disable=no-member
"""
Adjust continuous aggregate policies

Revision ID: 07d27793b5e2
Revises: 4a0f8c654257
Create Date: 2021-01-24 07:43:38.747557

"""
from alembic import op
from sqlalchemy.exc import ProgrammingError

# revision identifiers, used by Alembic.
revision = "07d27793b5e2"
down_revision = "4a0f8c654257"
branch_labels = None
depends_on = None


VIEWS = [
    {"name": "mv_nem_facility_power_5min", "interval": "30 minutes", "start_offset": "2 hours"},
    {"name": "mv_wem_facility_power_30min", "interval": "30 minutes", "start_offset": "2 hours"},
    {
        "name": "mv_interchange_power_nem_region",
        "interval": "30 minutes",
        "start_offset": "2 hours",
    },
    {
        "name": "mv_interchange_energy_nem_region",
        "interval": "30 minutes",
        "start_offset": "2 hours",
    },
    {"name": "mv_facility_energy_hour", "interval": "30 minutes", "start_offset": "2 hours"},
    {
        "name": "mv_balancing_summary_region_hour",
        "interval": "30 minutes",
        "start_offset": "2 hours",
    },
]

query_drop = "SELECT remove_continuous_aggregate_policy('{view_name}')"

query_create = """SELECT add_continuous_aggregate_policy('{view_name}',
    start_offset => INTERVAL '{start_offset}',
    end_offset => NULL,
    schedule_interval => INTERVAL '{schedule_size}');
"""


def upgrade() -> None:
    for v in VIEWS:
        q_drop = query_drop.format(view_name=v["name"])
        q_create = query_create.format(
            view_name=v["name"], schedule_size=v["interval"], start_offset=v["start_offset"]
        )

        with op.get_context().autocommit_block():
            try:
                op.execute(q_drop)
            except ProgrammingError:
                pass

        with op.get_context().autocommit_block():
            op.execute(q_create)


def downgrade() -> None:
    pass



import json
import logging
from itertools import groupby
from typing import List, Optional

from openpyxl import load_workbook

from opennem.core.dispatch_type import parse_dispatch_type
from opennem.core.facility.fueltechs import parse_facility_fueltech
from opennem.core.facilitystatus import parse_facility_status
from opennem.core.fueltechs import lookup_fueltech
from opennem.core.loader import PROJECT_DATA_PATH, load_data
from opennem.core.normalizers import (
    clean_capacity,
    normalize_duid,
    station_name_cleaner,
)
from opennem.core.stations.station_code_from_duids import (
    station_code_from_duids,
)
from opennem.core.stations.station_name_code_map import station_name_code_map
from opennem.core.unit_codes import get_unit_code
from opennem.core.unit_parser import parse_unit_duid
from opennem.exporter.encoders import OpenNEMJSONEncoder
from opennem.schema.stations import StationSet

logger = logging.getLogger("opennem.importer.mms")

participant_keys = ["name", "abn"]

FACILITY_KEYS = [
    "participant",
    "station_name",
    "region",
    "dispatch_type",
    "category",
    "classification",
    "fuel_source_primary",
    "fuel_source_descriptor",
    "tech_primary",
    "tech_primary_descriptor",
    "unit_no",
    "unit_size",
    "aggreagation",
    "duid",
    "reg_cap",
    "max_cap",
    "max_roc",
]


def lookup_station_code(
    duids: List[str], station_name: str, station_code_map: dict
) -> Optional[str]:

    station_code = None
    station_name = station_name.strip()

    for facility_duid in duids:
        if facility_duid in station_code_map:
            station_code = station_code_map[facility_duid]
            continue

    if not station_code:
        station_code = station_name_code_map(station_name)

    if not station_code:
        station_code = station_code_from_duids(duids)

    if not station_code:
        print(
            "Coult not get a station code for {}: {}".format(
                station_name, ",".join(duids)
            )
        )

    return station_code


def rel_grouper(records, station_code_map):
    records_parsed = []

    for _id, i in enumerate(records, start=2000):
        name = station_name_cleaner(i["station_name"])
        duid = normalize_duid(i["duid"])
        unit = parse_unit_duid(i["unit_no"], duid)
        fueltech = lookup_fueltech(
            i["fuel_source_primary"],
            i["fuel_source_descriptor"],
            i["tech_primary"],
            i["tech_primary_descriptor"],
            i["dispatch_type"],
        )
        station_code = lookup_station_code(
            [duid], i["station_name"], station_code_map
        )

        records_parsed.append(
            {
                "name": name,
                "code": duid,
                "status": parse_facility_status("operating"),
                "station_code": station_code,
                "network_region": i["region"].strip(),
                "network_name": i["station_name"].strip(),
                "unit_size": clean_capacity(i["unit_size"]),
                "unit_code": get_unit_code(unit, duid, name),
                "dispatch_type": parse_dispatch_type(i["dispatch_type"]),
                "fueltech": parse_facility_fueltech(fueltech),
                "capacity_registered": clean_capacity(i["reg_cap"]),
                "capacity_maximum": clean_capacity(i["max_cap"]),
            }
        )

    grouped_records = {}

    for key, v in groupby(records_parsed, key=lambda v: v["station_code"]):

        # key = k[1
        if key not in grouped_records:
            grouped_records[key] = []

        grouped_records[key] += list(v)

    coded_records = {}
    _id = 2000

    for station_code, rel in grouped_records.items():
        station_name = rel[0]["network_name"]

        if station_code in coded_records:
            raise Exception(
                "Code conflict: {}. {} {}".format(
                    station_code, station_name, coded_records[station_code]
                )
            )

        if not station_code:
            raise Exception("Unmapped station: {}".format(rel))

        coded_records[station_code] = {
            "name": station_name_cleaner(station_name),
            "network_name": station_name,
            "code": station_code,
            "id": _id,
            "facilities": rel,
        }

        _id += 1

    return coded_records


def load_rel():
    aemo_path = PROJECT_DATA_PATH / "aemo" / "nem_rel_202006.xlsx"

    if not aemo_path.is_file():
        raise Exception("Not found: {}".format(aemo_path))

    wb = load_workbook(aemo_path, data_only=True)

    generator_ws = wb["Generators and Scheduled Loads"]
    # participant_ws = wb.get_sheet_by_name("Registered Participants")

    generators = []

    for row in generator_ws.iter_rows(min_row=2, values_only=True):
        generators.append(
            dict(
                zip(
                    FACILITY_KEYS,
                    list(row[: len(FACILITY_KEYS)]),
                )
            )
        )

    return generators


def rel_import():
    mms_duid_station_map = load_data("mms_duid_station_map.json", True)

    nem_rel = load_rel()
    nem_rel = rel_grouper(nem_rel, mms_duid_station_map)

    rel = StationSet()

    for s in nem_rel.values():
        rel.add_dict(s)

    return rel


def rel_export():
    nem_rel = rel_import()

    with open("data/rel.json", "w") as fh:
        fh.write(nem_rel.json(indent=4))

    logger.info("Wrote {} records".format(nem_rel.length))


if __name__ == "__main__":
    rel_export()



from opennem.core.loader import load_data

FACILITY_DUID_MAP = load_data("facility_duid_map.json")


def facility_duid_map(duid: str) -> str:
    """
        Maps a DUID to a Facility code

    """
    if not type(FACILITY_DUID_MAP) is dict:
        raise Exception("Facility duid map invalid data type")

    if duid in FACILITY_DUID_MAP:
        return FACILITY_DUID_MAP[duid]

    return duid


def duid_is_retired(duid: str) -> bool:
    return duid in FACILITY_DUID_MAP



#!/usr/bin/env python
"""
Srapyd control methods

"""
import logging
from typing import Any, Dict, List
from urllib.parse import urljoin

from opennem.settings import settings
from opennem.utils.http import http
from opennem.utils.scrapy import get_spiders

logger = logging.getLogger("scrapyd.client")


def get_jobs() -> Dict[str, Any]:
    job_url = urljoin(
        settings.scrapyd_url,
        "listjobs.json?project={}".format(settings.scrapyd_project_name),
    )

    jobs = http.get(job_url).json()

    return jobs


def job_cancel(id: str) -> bool:
    cancel_job_url = urljoin(settings.scrapyd_url, "cancel.json")

    r = http.post(cancel_job_url, data={"project": "opennem", "job": id})

    resp = r.json()

    if resp["status"] == "error":
        logger.error("Error: {}".format(resp["message"]))
        return False

    logger.info("Cancelled job: {}".format(resp["jobid"]))

    return True


def job_schedule(spider_name: str) -> bool:
    schedule_url = urljoin(settings.scrapyd_url, "schedule.json")

    try:
        r = http.post(schedule_url, data={"project": "opennem", "spider": spider_name})
    except Exception as e:
        logger.error("Error getting {}: {}".format(schedule_url, e))
        return False

    if not r.ok:
        logger.error("Error: {}".format(r.status_code))
        return False

    resp = r.json()

    if resp["status"] == "error":
        logger.error("Error: {}".format(resp["message"]))
        return False

    logger.info("Queued spider {} with task: {}".format(spider_name, resp["jobid"]))

    return True


def job_cancel_state(state: str = "pending") -> bool:
    jobs = get_jobs()

    if state not in jobs:
        logger.info("Invalid state or no jobs in state {}".format(state))
        return False

    pending_jobs = jobs[state]

    for job in pending_jobs:
        job_id = job["id"]
        logger.info("Cancelling {}".format(job_id))
        job_cancel(job_id)

    return True


def job_schedule_all(matches: str = None) -> List[str]:
    spiders = get_spiders()

    spider_scheduled = []

    for s in spiders:
        if matches and matches not in s:
            continue
        job_schedule(s)
        spider_scheduled.append(s)

    return spider_scheduled



"""
OpenNEM v2 schema loader
"""


from typing import Dict, List

from opennem.core.compat.schema import OpennemDataSetV2, OpennemDataV2


def load_statset_v2(statset: List[Dict]) -> List[OpennemDataV2]:
    return OpennemDataSetV2(data=[OpennemDataV2.parse_obj(i) for i in statset])



from typing import Any, List, Union

from dictalchemy.utils import asdict
from pydantic import BaseModel

from opennem.db.models.opennem import BaseModel as BaseDbModel


def compare_record_differs(
    subject: object, target: object, fields: Union[str, List[str]]
) -> bool:
    """
        Compare two objects or types returns True if they're the same

    """

    if not type(fields) is list:
        fields = [fields]

    comparator_results = []

    for field in fields:
        subject_value = field_getter(subject, field)
        target_value = field_getter(target, field)

        if target_value in [None, {}, ""]:
            return False

        result = subject_value != target_value

        comparator_results.append(result)

    return all(comparator_results)


def compare_field(subject: Any, target: Any) -> bool:
    pass


def field_getter(subject: object, field_name: str, default: any = None):
    """
        Gets a field value from a dict or object

    """
    value = None

    if type(subject) is dict and field_name in subject:
        value = subject.get(field_name, default)

    if hasattr(subject, field_name):
        value = getattr(subject, field_name, default)

    if isinstance(value, BaseModel):
        value: dict = value.dict()

    if isinstance(value, BaseDbModel):
        value: dict = asdict(
            value, exclude=["created_by", "created_at", "updated_at"]
        )

    if value:
        return value

    return default



# pylint: disable=no-member
"""Optimized energy_sum sql function

Revision ID: 10f1d9517df7
Revises: 458bf49bd781
Create Date: 2020-10-29 03:25:01.093194

"""
from alembic import op

from opennem.core.loader import load_data

# revision identifiers, used by Alembic.
revision = "10f1d9517df7"
down_revision = "458bf49bd781"
branch_labels = None
depends_on = None


def upgrade():
    ENERGY_FUNCTIONS = load_data(
        "energy_sum_v2.sql", from_fixture=True, content_type="utf-8"
    ).decode("utf-8")
    op.execute(ENERGY_FUNCTIONS)


def downgrade():
    pass



from typing import Optional

from opennem.core.loader import load_data

STATION_CODE_REMAP = {"SWAN_B": "SWANBANK", "SWAN_E": "SWANBANK_E"}


def map_old_station_names(station_code: str) -> str:
    if station_code in STATION_CODE_REMAP:
        return STATION_CODE_REMAP[station_code]

    return station_code


STATION_DUID_MAP = load_data("facility_duid_map.json")


def facility_map_station(
    duid: str, default_station: Optional[str] = None
) -> Optional[str]:
    """
        Maps duid -> station code

    """

    if not type(STATION_DUID_MAP) is dict:
        raise Exception("Station duid map invalid data type")

    if duid in STATION_DUID_MAP:
        return STATION_DUID_MAP[duid]

    if default_station:
        station_code = map_old_station_names(default_station)

    return station_code


def facility_has_station_remap(duid: str) -> bool:
    return duid in STATION_DUID_MAP



"""
OpenNEM Compare Data

Compare JSON outputs between OpenNEM versions to check for consistency
"""


import json

from opennem.utils.http import http


def compare_data() -> None:

    return None


if __name__ == "__main__":
    compare_data()



import logging
from datetime import datetime, timedelta

from opennem.monitors.aemo_wem_live_intervals import (
    get_aemo_wem_live_facility_intervals_recent_date,
)
from opennem.notifications.slack import slack_message
from opennem.schema.network import NetworkWEM
from opennem.settings import settings
from opennem.utils.dates import chop_microseconds

logger = logging.getLogger("opennem.monitors.aemo")


def aemo_wem_live_interval() -> bool:
    """
    Monitors the delay from the AEMO live scada data on the portal
    """
    network = NetworkWEM

    now_date = datetime.now().astimezone(network.get_timezone())

    live_most_recent = get_aemo_wem_live_facility_intervals_recent_date()

    live_delta = chop_microseconds(now_date - live_most_recent)

    logger.debug(
        "Live time: {},  delay: {}".format(live_most_recent, live_delta)
    )

    # @TODO move the minutes into settings
    if live_delta > timedelta(minutes=90):
        slack_message(
            "*WARNING*: AEMO Live intervals for WEM on {} curently delayed by {}\n\nAEMO feed most recent: {}".format(
                settings.env, live_delta, live_most_recent
            )
        )
        return True

    return False


if __name__ == "__main__":
    delay = aemo_wem_live_interval()
